"""
_adapters_v1.model_adapters.bottleneck
================================

.. autoclass:: tamm._adapters_v1.model_adapters.bottleneck.BottleneckModelAdapter
    :show-inheritance:
    :members:
"""

import logging as _logging
from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm._adapters_v1 import layer_annotations as _layer_annotations
from tamm._adapters_v1.adapted_layer import AdaptedLayer as _AdaptedLayer
from tamm._adapters_v1.layer_adapters.bottleneck import Bottleneck as _Bottleneck
from tamm._adapters_v1.layer_adapters.bottleneck import (
    init_bias_params as _init_bias_params,
)
from tamm._adapters_v1.layer_adapters.bottleneck import (
    init_weight_params as _init_weight_params,
)
from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_annotations import LayerAnnotation as _LayerAnnotation
from tamm._adapters_v1.layer_annotations import LinearProjection as _LinearProjection
from tamm._adapters_v1.layer_annotations import (
    get_layer_annotations as _get_layer_annotations,
)
from tamm._adapters_v1.model_adapters.model_adapter import AdapterSpec as _AdapterSpec
from tamm._adapters_v1.model_adapters.model_adapter import ModelAdapter as _ModelAdapter
from tamm.layers import activation as _activation

_logger = _logging.getLogger(__name__)


class BottleneckModelAdapter(_ModelAdapter):
    """
    A model adapter which inserts :class:`_BottleneckLayerAdapter`.
    Two adapter variants supported are:

    * Bottleneck adapter: See Neil Houlsby, et al. (2019), Parameter-Efficient
    Transfer Learning for NLP for details. Here the adapter is attached after
    the feed-forward layer within the transformer block. More details are
    available in https://arxiv.org/pdf/1902.00751.
    * Residual adapter: See Katrin Tomanek, et al. (2021), Residual Adapters
    for Parameter-Efficient ASR Adaptation to Atypical and Accented Speech.
    Here the adapter contains layer norm added to the bottleneck layer.
    More details are available in https://arxiv.org/pdf/2109.06952.

    Args:
        input_dim: Layer input dim.
        output_dim: Layer output dim.
        hidden_dim: Layer hidden dim.
        activation: Layer activation.
        share_adapter_weights: Enable weight sharing.
        share_adapter_bias: Enable bias sharing.
        dropout_p: Dropout rate.
        layer_norm_eps: Layer norm eps.
        layer_norm_bias: Layer norm bias.
        input_normalize: Apply layer norm to inputs.
    """

    def __init__(
        self,
        hidden_dim: int = 128,
        activation: _activation.ActivationSpecType = ("relu",),
        adapt_feed_forward_hidden_states: bool = True,
        adapt_feed_forward_outputs: bool = True,
        share_adapter_weights: bool = False,
        share_adapter_bias: bool = False,
        dropout_p: float = 0,
        layer_norm_eps: float = 1e-5,
        layer_norm_bias: bool = True,
        input_normalize: bool = True,
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ):
        super().__init__()
        self.hidden_dim = hidden_dim
        self.activation = activation
        self.adapt_feed_forward_hidden_states = adapt_feed_forward_hidden_states
        self.adapt_feed_forward_outputs = adapt_feed_forward_outputs
        self.share_adapter_weights = share_adapter_weights
        self.share_adapter_bias = share_adapter_bias
        self.dropout_p = dropout_p
        self.layer_norm_eps = layer_norm_eps
        self.layer_norm_bias = layer_norm_bias
        self.input_normalize = input_normalize
        self.device = device
        self.dtype = dtype

        # adapter shareable parameters
        self._hidden_states_a_weights: _Optional[_nn.Parameter] = None
        self._hidden_states_a_bias: _Optional[_nn.Parameter] = None
        self._hidden_states_b_weights: _Optional[_nn.Parameter] = None
        self._hidden_states_b_bias: _Optional[_nn.Parameter] = None
        self._outputs_a_weights: _Optional[_nn.Parameter] = None
        self._outputs_a_bias: _Optional[_nn.Parameter] = None
        self._outputs_b_weights: _Optional[_nn.Parameter] = None
        self._outputs_b_bias: _Optional[_nn.Parameter] = None

    def _get_bottlenck_layer_type(
        self, layer_type: _List[_LayerAnnotation]
    ) -> _Optional[_LinearProjection]:
        for lt in layer_type:
            matching_type = (
                self.adapt_feed_forward_hidden_states
                and isinstance(lt, _layer_annotations.FeedForwardHiddenTransform)
            ) or (
                self.adapt_feed_forward_outputs
                and isinstance(lt, _layer_annotations.FeedForwardOutputTransform)
            )
            if matching_type:
                return lt
        return None

    def _create_adapters_impl(self, model: _nn.Module) -> _AdapterSpec:
        layer_adapters = {}
        for name, submodule in model.named_modules(remove_duplicate=True):
            if isinstance(submodule, _AdaptedLayer):
                adapter = self._create_adapter_for_submodule(submodule)
                if adapter is not None:
                    layer_adapters[name] = adapter
        return _AdapterSpec(layer_adapters=layer_adapters)

    def _get_hidden_states_params(self, dim: int):
        if self.share_adapter_weights:
            if self._hidden_states_a_weights is None:
                self._hidden_states_a_weights = _init_weight_params(
                    self.hidden_dim, dim, device=self.device, dtype=self.dtype
                )
            if self._hidden_states_b_weights is None:
                self._hidden_states_b_weights = _init_weight_params(
                    dim, self.hidden_dim, device=self.device, dtype=self.dtype
                )

        if self.share_adapter_bias:
            if self._hidden_states_a_bias is None:
                self._hidden_states_a_bias = _init_bias_params(
                    self.hidden_dim,
                    self._hidden_states_a_weights,
                    device=self.device,
                    dtype=self.dtype,
                )
            if self._hidden_states_b_bias is None:
                self._hidden_states_b_bias = _init_bias_params(
                    dim,
                    self._hidden_states_b_weights,
                    device=self.device,
                    dtype=self.dtype,
                )

        return (
            self._hidden_states_a_weights,
            self._hidden_states_a_bias,
            self._hidden_states_b_weights,
            self._hidden_states_b_bias,
        )

    def _get_outputs_params(self, dim: int):
        if self.share_adapter_weights:
            if self._outputs_a_weights is None:
                self._outputs_a_weights = _init_weight_params(
                    self.hidden_dim, dim, device=self.device, dtype=self.dtype
                )
            if self._outputs_b_weights is None:
                self._outputs_b_weights = _init_weight_params(
                    dim, self.hidden_dim, device=self.device, dtype=self.dtype
                )

        if self.share_adapter_bias:
            if self._outputs_a_bias is None:
                self._outputs_a_bias = _init_bias_params(
                    self.hidden_dim,
                    self._outputs_a_weights,
                    device=self.device,
                    dtype=self.dtype,
                )
            if self._outputs_b_bias is None:
                self._outputs_b_bias = _init_bias_params(
                    dim, self._outputs_b_weights, device=self.device, dtype=self.dtype
                )

        return (
            self._outputs_a_weights,
            self._outputs_a_bias,
            self._outputs_b_weights,
            self._outputs_b_bias,
        )

    def _get_shared_params(self, layer_type: _LayerAnnotation, input_dim: int):
        if isinstance(layer_type, _layer_annotations.FeedForwardHiddenTransform):
            return self._get_hidden_states_params(input_dim)
        if isinstance(layer_type, _layer_annotations.FeedForwardOutputTransform):
            return self._get_outputs_params(input_dim)
        return (None, None, None, None)

    def _create_adapter_for_submodule(
        self, submodule: _nn.Module
    ) -> _Optional[_LayerAdapter]:
        wrapped_module = submodule.unwrap()
        layer_type = _get_layer_annotations(wrapped_module)
        bn_layer_type = self._get_bottlenck_layer_type(layer_type)
        if bn_layer_type:
            input_dim = getattr(wrapped_module, bn_layer_type.output_dim_field)
            a_weights, a_bias, b_weights, b_bias = self._get_shared_params(
                bn_layer_type, input_dim
            )

            return _Bottleneck(
                input_dim=input_dim,
                output_dim=input_dim,
                hidden_dim=self.hidden_dim,
                a_weights=a_weights,
                a_bias=a_bias,
                b_weights=b_weights,
                b_bias=b_bias,
                activation=_activation.create_activation_layer(self.activation),
                dropout_p=self.dropout_p,
                layer_norm_eps=self.layer_norm_eps,
                layer_norm_bias=self.layer_norm_bias,
                input_normalize=self.input_normalize,
                device=self.device,
                dtype=self.dtype,
            )
        return None



from tamm import _warnings as _tamm_warnings
from tamm.model_repo.publishing import PublishedModelConfig as _PublishedModelConfig


def _warn_deprecated_published_config(config: "_PublishedModelConfig"):
    if config.is_deprecated:
        msg = f"'{config.model_id}' model is deprecated."
        replacement_model_id = config.replacement_model_id
        if replacement_model_id:
            msg = msg + f" Please use '{replacement_model_id}' model instead."
        _tamm_warnings.deprecation(msg)



"""
Utilities for discovering and potentially importing plugins for |tamm|.
"""
import importlib
import logging as _logging
import sys
from typing import Any as _Any
from typing import Dict as _Dict

from tamm.context_vars import _plugin_init
from tamm.runtime_configuration import rc

if sys.version_info < (3, 10):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata


_logger = _logging.getLogger(__name__)


def execute_plugin_object_reference(object_ref: str, *, call_obj: bool = True) -> _Any:
    """
    Imports a plugin submodule/object and optionally calls it.

    Args:
        object_ref (:obj:`str`): A string of the format ``importable.module`` or
            ``importable.module:object.attr``.
        call_obj (:obj:`bool`): A flag that controls calling the referenced object.
            If the object is not callable, the flag has no effect.  Defaults to
            ``True``.

    Returns:
        The result of the object call if it is called, otherwise the object itself.
    """

    try:
        modname, separator, qualname = object_ref.partition(":")
        root_modname = modname.split(".", maxsplit=1)[0]

        with _plugin_init.plugin_initialization_context(root_modname):
            _logger.debug(f"Importing plugin module {modname}")
            obj = importlib.import_module(modname)
            _logger.debug(f"{modname} import complete")

            if separator:
                for attr in qualname.split("."):
                    obj = getattr(obj, attr)

            if call_obj and callable(obj):
                _logger.debug(f"Calling plugin object reference {object_ref}")
                result = obj()
                _logger.debug(f"{object_ref} call complete")
                return result

            return obj

    except Exception as e:
        _logger.debug(f"Caught exception during {object_ref} plugin handling: {e}")

    return None


def discover_plugin_extras(extras_name: str) -> importlib_metadata.EntryPoints:
    """
    Retrieves an iterable of entrypoints for a type of plugin extra.

    Args:
        extras_name: The entrypoint name for the extras.
    """
    return importlib_metadata.entry_points(
        name=extras_name, group=rc.PLUGINS_EXTRAS_ENTRYPOINT_GROUP
    )


def discover_and_retrieve_plugins() -> _Dict[str, str]:
    """
    Discovers plugin by entrypoints mechanism:
    https://packaging.python.org/en/latest/specifications/entry-points/

    All plugins with the plugins entrypoint group name with be discovered and
    returned.

    Returns:
        (:obj:`_Dict[str, str]`): Discovered plugins dictionary of the format
        ```
        {
            "plugin-name": "plugin_module"
        }
        ```
    """

    plugins: _Dict[str, str] = {}

    discovered_plugins = importlib_metadata.entry_points(
        group=rc.PLUGINS_ENTRYPOINT_GROUP
    )

    for plugin in discovered_plugins:
        plugins[plugin.name] = plugin.value
    return plugins


def import_named_plugin(plugin_module_name: str) -> None:
    """
    Import the specified plugin module.

    Args:
        plugin_module_name (:obj:`str`): plugin module to
            import a specific plugin.

    Raises:
        ValueError: If the plugin module name is not a string or not installed.
        ImportError: If there is an exception importing the plugin module.
    """

    if not isinstance(plugin_module_name, str):
        raise ValueError(
            f"plugin_module must be a str, got "
            f"type({type(plugin_module_name).__name__})"
        )

    plugins = discover_and_retrieve_plugins()
    if plugin_module_name not in plugins.values():
        raise ValueError(
            f"{plugin_module_name} is not in the list of installed plugins. "
            f"Ensure {plugin_module_name} plugin is installed."
        )
    try:
        importlib.import_module(plugin_module_name)
        _logger.debug(f"Successfully imported {plugin_module_name}.")
    except ImportError as e:
        raise ImportError(
            f"'{plugin_module_name}' is discovered but cannot be imported by tamm"
        ) from e


def import_discovered_plugins() -> None:
    """
    Retrieves plugins and attempts runtime import of all discovered plugins.

    Raises:
        ImportError: If there is an exception importing any plugin module.
    """

    plugins = discover_and_retrieve_plugins()

    for plugin_name, plugin_module in plugins.items():
        try:
            import_named_plugin(plugin_module)
        except (ImportError, ValueError) as e:
            _logger.debug("%s", e)
        except Exception as e:
            _logger.debug(
                "unknown exception occurred when importing plugin %s (%s). "
                "tamm will proceed anyways, please inform plugin owner and/or tamm team about this error: %s",
                plugin_name,
                plugin_module,
                str(e),
            )



"""
This module implements loss functions specific to mixture of experts layers.
"""


import torch as _torch

from tamm.layers.common import LayerMixin as _LayerMixin


class LoadBalanceLoss(_torch.nn.Module, _LayerMixin):
    """
    A differentiable load balancing loss for MoE routers.  For more details,
    see the `Switch Transformers paper <https://arxiv.org/abs/2101.03961>`__.
    This loss is normalized such that its value is ``1.0`` when the expert
    assignments and assignment probabilities are both uniform.
    """

    def __init__(self):
        super().__init__()

    def forward(
        self,
        expert_assignments: _torch.Tensor,
        assignment_probabilities: _torch.Tensor,
    ):
        """
        Computes and returns the loss value.

        Args:
            expert_assignments (:obj:`torch.Tensor`): An integer tensor that
                maps tokens to chosen experts.  The expected shape is
                ``(*batch_shape, seq_len, num_experts_per_token)``.
            assignment_probabilities (:obj:`torch.Tensor`): The output of the
                router's softmax op.  The expected shape is
                ``(*batch_shape, seq_len, num_experts)``.
        """

        num_experts = assignment_probabilities.size(-1)

        # pylint: disable-next=not-callable
        one_hot_assignments = _torch.nn.functional.one_hot(
            expert_assignments, num_classes=num_experts
        )  # (T, B, S, K, E)
        counts = one_hot_assignments.count_nonzero(dim=(-3, -2))  # (T, B, E)
        denom = one_hot_assignments.shape[-3:-1].numel()  # S * K

        soft_freqs = assignment_probabilities.mean(-2)  # (T, B, E)

        return _torch.mean(counts * soft_freqs) * (num_experts**2 / denom)


class RouterZLoss(_torch.nn.Module, _LayerMixin):
    """
    A loss function that penalizes large logits in MoE routers.  For more
    details, see the `ST-MoE paper <https://arxiv.org/abs/2202.08906>`__.
    """

    def __init__(self):
        super().__init__()

    def forward(self, logits: _torch.Tensor) -> _torch.Tensor:
        """
        Computes and returns the loss value.

        Args:
            logits (:obj:`torch.Tensor`): The input to the router's
                softmax op.  The expected shape is
                ``(*batch_shape, seq_len, num_experts)``.
        """

        tensor = _torch.logsumexp(logits, dim=-1)
        return tensor.square().mean()



"""
token_generators.greedy
^^^^^^^^^^^^^^^^^^^^^^^
"""

from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch
import torch.nn.functional as _F
from torch import nn as _nn

from tamm.generation.token_generators._vanilla import _VanillaTokenGenerator
from tamm.generation.token_generators.base import (
    TokenGeneratorOutput as _TokenGeneratorOutput,
)
from tamm.generation.token_generators.next_token_logic import (
    NextTokenLogic as _NextTokenLogic,
)
from tamm.generation.token_generators.next_token_logic import (
    NextTokenLogicOutput as _NextTokenLogicOutput,
)
from tamm.generation.token_generators.utils import KVCacheType as _KVCacheType
from tamm.utils.torch_utils import (
    torch_exportable_dataclass as _torch_exportable_dataclass,
)


class GreedyNextTokenLogic(_NextTokenLogic):
    """
    Generates next token as the argmax of the model's output logits.
    """

    def compute_probabilities(self, logits: _torch.Tensor) -> _torch.Tensor:
        """
        Returns a one-hot probability distribution with value ``1`` at the index
        corresponding to argmax of logits.
        """
        # pylint: disable=not-callable
        indices = logits.argmax(dim=-1)
        one_hot = _F.one_hot(indices, num_classes=logits.size(-1))
        return one_hot.to(logits.dtype)

    def __call__(
        self,
        logits: _torch.Tensor,
        return_probabilities: bool = False,
        rng: _Optional[_torch.Generator] = None,
    ) -> _NextTokenLogicOutput:
        if not return_probabilities:
            return _NextTokenLogicOutput(token_ids=logits.argmax(dim=-1, keepdim=True))
        probabilities = self.compute_probabilities(logits)
        return _NextTokenLogicOutput(
            token_ids=probabilities.argmax(dim=-1, keepdim=True),
            probabilities=probabilities,
        )


@_torch_exportable_dataclass
class GreedyTokenGeneratorOutput(_TokenGeneratorOutput):
    """
    Output generated from :py:class:`GreedyTokenGenerator`, consisting of token ids and
    corresponding logits, if that option is enabled.

    Args:
        token_ids (:py:class:`_torch.Tensor`): Token ids produced by the greedy generator.
          Does not include ``input_ids``.
        logits (optional :py:class:`_torch.Tensor`): Logits corresponding to the generated token ids.
        probabilities (optional :py:class:`_torch.Tensor`): Probability distributions corresponsing to
         the token ids.
    """

    token_ids: _torch.Tensor
    logits: _Optional[_torch.Tensor] = None
    probabilities: _Optional[_torch.Tensor] = None


class GreedyTokenGenerator(_VanillaTokenGenerator):
    """
    Token generator that uses greedy decoding strategy.

    Args:
        model (:obj:`nn.Module`): The model instance.
        eos_id (:obj:`Union[int, torch.Tensor, List[torch.Tensor]]`): End of sequence token ID(s).
        pad_id (:obj:`Union[int, torch.Tensor]`): Padding token ID.
        cache_type (:obj:`KVCacheType`, optional): Type of KV cache to use. Defaults to KVCacheType.V0.
        cache_dtype (:obj:`torch.dtype`, optional): Data type for KV cache. Defaults to None.
        output_logits (:obj:`bool`): When set to ``True``, :py:meth:`generate` returns logits
          corresponding to the generated ``token_ids``.
        output_probabilities (:obj:`bool`): When set to ``True``, :py:meth:`generate` returns probability
          distributions corresponding to the generated ``token_ids``.
    """

    def __init__(
        self,
        model: _nn.Module,
        eos_id: _Union[int, _torch.Tensor, _List[_torch.Tensor]],
        pad_id: _Union[int, _torch.Tensor],
        cache_type: _Optional[_KVCacheType] = _KVCacheType.V0,
        cache_dtype: _Optional[_torch.dtype] = None,
        output_logits: bool = False,
        output_probabilities: bool = False,
    ):
        next_token_logic = GreedyNextTokenLogic()
        # pylint: disable=duplicate-code
        super().__init__(
            model=model,
            next_token_logic=next_token_logic,
            eos_id=eos_id,
            pad_id=pad_id,
            cache_type=cache_type,
            cache_dtype=cache_dtype,
            output_logits=output_logits,
            output_probabilities=output_probabilities,
        )

    @_torch.inference_mode()
    def generate(
        self,
        input_ids: _torch.Tensor,
        max_new_tokens: int,
        seed: _Optional[int] = None,
    ) -> GreedyTokenGeneratorOutput:
        token_ids, return_logits, return_probs = self._generate(
            input_ids, max_new_tokens, seed
        )
        return GreedyTokenGeneratorOutput(
            token_ids=token_ids, logits=return_logits, probabilities=return_probs
        )



from tamm._cli._listers._base import BaseLister
from tamm._cli._listers._helpers import (
    PrettyPrintListedEntities as _PrettyPrintListedEntities,
)
from tamm.model_repo import get_model_repo_lazy


class ModelsLister(BaseLister):
    name = "models"

    # pylint: disable=redefined-builtin
    def __init__(self, all, long, wide, show_deprecated):
        super().__init__(all=all, long=long, wide=wide, show_deprecated=show_deprecated)
        kwargs = {"certified_tags": None} if self.all else {}
        self._model_repo = get_model_repo_lazy(**kwargs)
        self._listing_function = self._model_repo.list_models

    def print(self):
        _PrettyPrintListedEntities(self._listing_function).print(
            include_descriptions=self.long,
            filter_deprecated=not self.all,
            wide=self.wide,
        )


class AdaptedModelsLister(ModelsLister):
    name = "adapted-models"

    # pylint: disable=redefined-builtin
    def __init__(self, all, long, wide, show_deprecated):
        super().__init__(all=all, long=long, wide=wide, show_deprecated=show_deprecated)
        self._listing_function = self._model_repo.list_adapted_models



from typing import ContextManager

from tamm.context_vars import pretrained_flag_context
from tamm.layers.common.context_hooks.base import _BaseContextHook
from tamm.typing import LenientOptionalBool
from tamm.utils import OptionalBool as _OptionalBool


class PretrainedContextHook(_BaseContextHook):
    def __init__(self, pretrained: "LenientOptionalBool" = _OptionalBool.NOTSET):
        self.pretrained = _OptionalBool(pretrained)

    def is_null(self) -> bool:
        return self.pretrained == _OptionalBool.NOTSET

    def _get_context_manager(self) -> "ContextManager":
        return pretrained_flag_context(self.pretrained)



import abc as _abc
from typing import Optional as _Optional

import torch as _torch

from tamm.utils.torch_utils import (
    torch_exportable_dataclass as _torch_exportable_dataclass,
)


@_torch_exportable_dataclass
class NextTokenLogicOutput:
    """
    Output generated from :py:class:`NextTokenLogic`, consisting of token ids and
    corresponding probabilities, if that option is enabled.

    Args:
        token_ids (:py:class:`_torch.Tensor`): New token ids produced by the next token logic.
          token_ids have the same number of dimensions as the input logits, i.e., if the
          input logits have shape [bs x seq_len x vocab_size], the token_ids would have shape
          [bs x seq_len x 1].
        probabilities (optional :py:class:`_torch.Tensor`): Probability distributions corresponsing to
         the generated token ids.
    """

    token_ids: _torch.Tensor
    probabilities: _Optional[_torch.Tensor] = None


class NextTokenLogic(_abc.ABC):
    """
    An abstract class for implementing logic for generating next token given a
    model's output logits for the current input_ids.
    """

    @_abc.abstractmethod
    def compute_probabilities(self, logits: _torch.Tensor) -> _torch.Tensor:
        """
        Computes a probability distribution from a model's logits.
        """

    @_abc.abstractmethod
    def __call__(
        self,
        logits: _torch.Tensor,
        return_probabilities: bool = False,
        rng: _Optional[_torch.Generator] = None,
    ) -> NextTokenLogicOutput:
        """
        Compute next token given the model's logits. Also return probabilities
        when return_probabilities is set to ``True``.
        """



"""
layers.flash_attention
^^^^^^^^^^^^^^^^^^^^^^

This submodule contains helper functions for using
`FlashAttention <https://github.com/Dao-AILab/flash-attention>`__ with |tamm| layers.

.. autoclass:: tamm.layers.flash_attention.FlashAttentionMode
    :members:

.. autofunction:: tamm.layers.flash_attention.compute_flash_attention_options
"""

import enum as _enum
import warnings as _warnings
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Tuple as _Tuple

import torch as _torch
from packaging.version import parse

from tamm import _helpers
from tamm._helpers import case_insensitive_lookup
from tamm.runtime_configuration import rc as _rc

# This is to allow full-graph compilation
_IS_FLASH_ATTENTION_INITIALIZED = False
_IS_FLASH_ATTENTION_AVAILABLE = False

# torch.exp(torch.tensor(-1e4, device="cuda", dtype=torch.float16)) -> 0
# torch.exp(torch.tensor(-1e4, device="cuda", dtype=torch.bfloat16)) -> 0
# torch.exp(torch.tensor(-1e4, device="cuda", dtype=torch.float32)) -> 0
SMALL_NEGATIVE_VALUE = -1e4

FLASH_SUPPORTED_DTYPES = (_torch.float16, _torch.bfloat16)


def is_flash_attention_available() -> bool:
    # pylint:disable=global-statement
    global _IS_FLASH_ATTENTION_INITIALIZED
    global _IS_FLASH_ATTENTION_AVAILABLE

    if not _IS_FLASH_ATTENTION_INITIALIZED:
        _IS_FLASH_ATTENTION_INITIALIZED = True

        # Don't even try if hardware or CUDA do not support it
        if (
            # if we are not running on GPUs
            not _torch.cuda.is_available()
            # or GPU is too old
            or _torch.cuda.get_device_capability()[0] < 8
            # or CUDA is too old
            or (
                _torch.version.cuda is not None
                and parse(_torch.version.cuda) <= parse("11.6")
            )
        ):
            return False

        try:
            # pylint:disable=import-outside-toplevel,unused-import
            import flash_attn  # noqa=F401

            # Need >= 2.4.0 for ALiBi support
            flash_version = parse(flash_attn.__version__)
            if flash_version < parse("2.4.0"):
                _warnings.warn(
                    "Flash attention version is too old,"
                    f" expecting >= 2.4.0, got {flash_version}"
                )
            else:
                _IS_FLASH_ATTENTION_AVAILABLE = True

        except ModuleNotFoundError as e:
            _helpers.log_exception(e, "importing flash_attn")
            _warnings.warn(
                "Failed to import flash-attn for Flash attention. Using "
                "flash attention may lead to significantly faster training. "
                "Please refer to tamm-scripts/install_flash_attn.sh for instructions."
            )

    return _IS_FLASH_ATTENTION_AVAILABLE


_MAX_NUM_SEGMENTS = int(_rc.flash_attn_max_num_segments)
if _MAX_NUM_SEGMENTS % 8 != 0:
    raise ValueError(
        "`TAMM_FLASH_ATTN_MAX_NUM_SEGMENTS` has to be a multiple of 8,"
        f" got {_MAX_NUM_SEGMENTS}"
    )


def compute_flash_attention_biases(
    *,
    segment_ids: _torch.Tensor,
    kv_segment_ids: _torch.Tensor,
    biases_dtype: _torch.dtype,
) -> _Tuple[_torch.Tensor, _torch.Tensor]:
    """
    Returns low rank Q, K biases for Flash attention.
    """
    # In the following lines of code, we adopt a method to add custom
    # padding (encoded in the segment_ids) without building a high-rank bias
    # matrix and adding that to the attention score matrix.
    # We add 8 extra values in the hidden dimension of the Q, K instead.
    # We needed to add 8 instead of 1 because of flash_attn_func's constraint
    # requiring the hidden dimension to be divisible by 8. For each segment,
    # we add one value, so there can't be more than 8 segments.
    # Detail below:
    #
    # Let Q, K, V : seqlen x dim matrices.
    # Consider the following table:
    #                                              Target bias term value
    # padding row in Q,     padding col in K            -inf
    # padding row in Q,     non-padding col in K        -inf
    # non-padding row in Q, padding col in K            -inf
    # non-padding row in Q, non-padding col in K          0
    #
    # In order to make the above happen, we create biases as follows:
    # For every padding row in Q, add a 2 bias term
    # For every padding row in K, add a -inf bias term
    # For every non-padding row in Q, add a 0 bias term
    # For every non-padding row in K, add a 0 bias term
    #
    # Note that in practice, because 0 * inf = nan, we use the smallest
    # possible negative value as -smallest_value. 2 * (-smallest_value) = -inf.

    if not segment_ids.ndim == kv_segment_ids.ndim == 2:
        raise ValueError("This function only supports 2-dimensional segment ids")

    range_n = _torch.arange(
        _MAX_NUM_SEGMENTS, device=segment_ids.device, dtype=_torch.int32
    )
    mask_query = range_n[None, :] == segment_ids[:, :, None]
    mask_key = range_n[None, :] == kv_segment_ids[:, :, None]

    zero = _torch.zeros(1, dtype=biases_dtype, device=segment_ids.device)
    low_rank_mask_query = _torch.where(mask_query, 1.0, zero)
    low_rank_mask_key = _torch.where(mask_key, zero, SMALL_NEGATIVE_VALUE)

    low_rank_mask_query = low_rank_mask_query[:, :, None]  # add heads dimension
    low_rank_mask_key = low_rank_mask_key[:, :, None]

    return low_rank_mask_query, low_rank_mask_key


def _qkv_flash_attention_compatible(
    query: _torch.Tensor, key: _torch.Tensor, value: _torch.Tensor
) -> bool:
    return all(
        _.device.type == "cuda" and _.dtype in FLASH_SUPPORTED_DTYPES
        for _ in [query, key, value]
    )


def _maybe_pad_qkv(
    *,
    query: _torch.Tensor,
    key: _torch.Tensor,
    value: _torch.Tensor,
    query_bias: _torch.Tensor,
    key_bias: _torch.Tensor,
    value_padding: _torch.Tensor,
) -> _Tuple[bool, _torch.Tensor, _torch.Tensor, _torch.Tensor]:
    is_padded = False

    if query_bias is not None or key_bias is not None or value_padding is not None:
        is_padded = True

        # All paddings are in shape
        # (batch_size, seq_len, num_heads, max_num_segments)
        query_bias = query_bias.expand(*query.shape[:-1], -1)
        query = _torch.cat([query, query_bias], dim=-1)

        key_bias = key_bias.expand(*key.shape[:-1], -1)
        key = _torch.cat([key, key_bias], dim=-1)

        value_padding = value_padding.expand(*value.shape[:-1], -1)
        value = _torch.cat([value, value_padding], dim=-1)

    return is_padded, query, key, value


def _flash_attn_func(*args, **kwargs):
    # pylint:disable=import-error,import-outside-toplevel
    from flash_attn import flash_attn_func

    return flash_attn_func(*args, **kwargs)


def flash_attn_func_with_low_rank_mask(
    *,
    query: _torch.Tensor,
    key: _torch.Tensor,
    value: _torch.Tensor,
    dropout_p: _Optional[float] = 0.0,
    scale: _Optional[float] = None,
    segment_ids: _Optional[_torch.Tensor] = None,
    alibi_slopes: _Optional[_torch.Tensor] = None,
    window_size=(-1, -1),
    query_bias: _Optional[_torch.Tensor] = None,
    key_bias: _Optional[_torch.Tensor] = None,
    value_padding: _Optional[_torch.Tensor] = None,
    is_causal: bool = False,
) -> _torch.Tensor:
    if query.dtype != key.dtype != value.dtype:
        raise TypeError(
            "Flash attention requires Q, K, V to have the same dtype"
            " -- torch.float16 or torch.bfloat16."
        )

    # pylint:disable=protected-access
    if segment_ids.device.type != "mps":
        _torch._assert_async(segment_ids.max() < _MAX_NUM_SEGMENTS)

    dim_per_head = key.shape[-1]
    if scale is None:
        scale = dim_per_head**-0.5

    is_padded, query, key, value = _maybe_pad_qkv(
        query=query,
        key=key,
        value=value,
        query_bias=query_bias,
        key_bias=key_bias,
        value_padding=value_padding,
    )

    if alibi_slopes is not None:
        alibi_slopes = alibi_slopes.type(_torch.float32)  # flash requires alibi in fp32

    output = _flash_attn_func(
        q=query,
        k=key,
        v=value,
        dropout_p=dropout_p,
        softmax_scale=scale,
        window_size=window_size,
        causal=is_causal,
        alibi_slopes=alibi_slopes,
    )

    if is_padded:
        return output[..., :dim_per_head]

    return output


class FlashAttentionMode(str, _enum.Enum):
    """An :class:`Enum` for controlling the FlashAttention integration."""

    #: Let |tamm| determine the mode.
    AUTO = "AUTO"

    #: Turn off the integration with FlashAttention.
    NONE = "NONE"

    @classmethod
    def _missing_(cls, value):
        return case_insensitive_lookup(cls, value)


def compute_flash_attention_options(
    *,
    mode: FlashAttentionMode,
    segment_ids: _torch.Tensor,
    kv_segment_ids: _torch.Tensor,
    compute_dtype: _torch.dtype,
    is_causal: bool = False,
    attention_mask: _Optional[_torch.Tensor] = None,
    segment_mask: _Optional[_torch.Tensor] = None,
) -> _Dict[str, _Any]:
    # pylint: disable=unused-argument

    if not is_flash_attention_available():
        return None

    if mode is FlashAttentionMode.NONE:
        return None

    query_bias, key_bias = compute_flash_attention_biases(
        segment_ids=segment_ids,
        kv_segment_ids=kv_segment_ids,
        biases_dtype=compute_dtype,
    )

    return {
        "segment_ids": kv_segment_ids,
        "query_bias": query_bias,
        "key_bias": key_bias,
        "value_padding": _torch.zeros_like(key_bias),
        "is_causal": is_causal,
    }



"""
Similar to PyTorch `default <https://pytorch.org/docs/stable/tensor_attributes.html#>`_
``device`` which can be accessed via
:meth:`torch.get_default_device` and :meth:`torch.get_default_device`, |tamm| has
the following global defaults which may affect module creation.

* ``pretrained`` (default: ``False``)

Controls whether a model will be loaded with pretrained weights.

.. autofunction:: tamm.get_default_pretrained_flag
.. autofunction:: tamm.set_default_pretrained_flag
.. autofunction:: tamm.pretrained_flag_context

* ``freeze_params`` (default: :class:`OptionalBool`.NOTSET)

Controls whether module parameters are frozen when created. Default is NOTSET meaning
the builder will not change the state of ``require_grad`` (i.e., ``not freeze_params``)

.. autofunction:: tamm.get_default_freeze_params_flag
.. autofunction:: tamm.set_default_freeze_params_flag
.. autofunction:: tamm.freeze_params_flag_context

* ``first_token_generation`` (default: ``False``)

Indicates whether the model is currently generating the first token in a sequence.

.. autofunction:: tamm.get_first_token_generation_flag
.. autofunction:: tamm.first_token_generation_context

Context Resolvers
^^^^^^^^^^^^^^^^^
.. autofunction:: tamm.resolve_device
.. autofunction:: tamm.resolve_pretrained_flag
.. autofunction:: tamm.resolve_freeze_params

"""

from tamm.context_vars._first_token_generation import (
    first_token_generation_context,
    get_first_token_generation_flag,
)
from tamm.context_vars._freeze_params import (
    freeze_params_flag_context,
    get_default_freeze_params_flag,
    set_default_freeze_params_flag,
)
from tamm.context_vars._model_build_device import (
    get_model_build_device,
    model_build_device_context,
)
from tamm.context_vars._pretrained import (
    get_default_pretrained_flag,
    pretrained_flag_context,
    set_default_pretrained_flag,
)
from tamm.context_vars._resolvers import (
    resolve_device,
    resolve_freeze_params,
    resolve_pretrained_flag,
)

__all__ = [
    "get_model_build_device",
    "model_build_device_context",
    "resolve_device",
    "resolve_pretrained_flag",
    "resolve_freeze_params",
    "get_default_pretrained_flag",
    "set_default_pretrained_flag",
    "pretrained_flag_context",
    "get_default_freeze_params_flag",
    "set_default_freeze_params_flag",
    "freeze_params_flag_context",
    "get_first_token_generation_flag",
    "first_token_generation_context",
]



"""
utils.partial
=============

This module implements the :class:`.DataclassedPartial` concept.  This is a base class
for config-like classes in |tamm|, such as :class:`.ModelConfig` and
:class:`.LayerBuilder`.

The purpose of a :class:`.DataclassedPartial` is to organize the configuration of a
callable's arguments without actually calling the callable.  The callable is typically
a complex class, such as a |tamm| model.  This separation of configuration from
invocation is important for many use cases.  For example, model developers often
publish a serialized :obj:`.ModelConfig` so that other |tamm| users may easily recreate
the same models.

:class:`.DataclassedPartial` is very similar to Python's :func:`functools.partial`
concept.  The main difference is that :class:`.DataclassedPartial` provides a
dataclass-style user experience for updating arguments to the target callable.  Unlike
the basic partial, we also create an explicit subclass of :class:`.DataclassedPartial`
for each target callable.

:class:`.DataclassedPartial` provides a Pythonic solution for object configuration.
It eliminates much of the config-related boilerplate found in packages such as Hugging
Face, since :class:`.DataclassedPartial` automatically derives each config class from
the function or class being configured.  This also preserves the regular user experience
of the callable---if desired, users may still instantiate objects or call functions
directly with normal arguments, avoiding config objects altogether.

**Examples:**

Configure a class instance prior to instantiation:

.. code-block:: python

    class Square:
        def __init__(self, side_length, color=None):
            self.side_length = side_length
            self.color = color

        def __str__(self):
            return f"Square(side_length={self.side_length}, color={self.color})"


    SquareConfig = DataclassedPartial.create_subclass(Square, name="SquareConfig")
    config = SquareConfig(8)
    config.color = "blue"
    square = config()
    print(square)  # 'Square(side_length=8, color=blue)'


Configure a class instance when the class initializer takes varargs and kwargs:

.. code-block:: python

    class Polygon:
        def __init__(self, *side_lengths, **metadata):
            self.side_lengths = side_lengths
            self.metadata = metadata

        def __str__(self):
            return (
                f"Polygon(side_lengths={self.side_lengths}, metadata={self.metadata})"
            )


    PolygonConfig = DataclassedPartial.create_subclass(Polygon, name="PolygonConfig")
    config = PolygonConfig(color="purple")
    config.side_lengths = [3, 4, 5]
    config.metadata["weight"] = 7
    polygon = config()
    print(polygon)
    # 'Polygon(side_lengths=(3, 4, 5), metadata={'color': 'purple', 'weight': 7})'


Use hooks to modify the result of a partial:

.. code-block:: python

        class TensorHolder:
            def __init__(self, value: float):
                self.tensor = torch.tensor(value, dtype=torch.float)

            def __str__(self):
                return f"TensorHolder(tensor={repr(self.tensor)})"


        TensorHolderPartial = DataclassedPartial.create_subclass(
            TensorHolder, name="TensorHolderPartial"
        )
        partial = TensorHolderPartial(value=0)

        print(partial())  # "TensorHolder(tensor=tensor(0.))"


        def use_gpu_context_hook():
            ""\"
            A context hook that adds a ``with torch.device("cuda"):`` context when
            creating the TensorHolder in ``partial.__call__()``.  Within the context,
            this sets the ``torch`` default device to a GPU.
            ""\"
            return torch.device("cuda")


        def return_tensor_post_hook(result):
            ""\"
            A post hook that causes the partial to return the ``tensor`` from the
            TensorHolder rather than the TensorHolder itself.
            ""\"
            return result.tensor


        def add_7_post_hook(tensor):
            ""\"A post hook that adds 7 to the resulting tensor.""\"
            tensor.add_(7)


        partial.context_hooks.register(use_gpu_context_hook)
        partial.post_hooks.register(return_tensor_post_hook)
        partial.post_hooks.register(add_7_post_hook)  # hooks fire in registration order
        print(partial())  # "tensor(7., device='cuda:0')"


Configure only some arguments of a function (similar to :func:`functools.partial`):

.. code-block:: python

    def y(x, m=1, b=0):
        return m * x + b

    PartialY = DataclassedPartial.create_subclass(y, name="PartialY")
    partial_y = PartialY(m=2, b=2)
    print(partial_y(0))  # '2'
    print(partial_y(2))  # '6'
    partial_y.m = 0
    print(partial_y(2))  # '2'


.. autoclass:: tamm.utils.partial.DataclassedPartial
    :members:
    :special-members: __call__, __init_subclass__

.. autoclass:: tamm.utils.partial.PartialPostHooks
    :members:
    :inherited-members:

.. autoclass:: tamm.utils.partial.PartialContextHooks
    :members:
    :inherited-members:

.. autoclass:: tamm.utils.partial.PartialHookHandle
    :members:
"""

import contextlib as _contextlib
import dataclasses as _dataclasses
import inspect as _inspect
import logging as _logging
import textwrap as _textwrap
import typing as _typing
import weakref as _weakref
from typing import Any as _Any
from typing import Callable as _Callable
from typing import Optional as _Optional

from tamm import _helpers

_logger = _logging.getLogger(__name__)


class DataclassedPartial:
    """
    A base class for configs.  For each subclass of :class:`.DataclassedPartial`,
    the constructor's signature matches the signature of the target callable.
    """

    _CALLABLE: _Optional[_Callable[..., _Any]] = None

    def __init_subclass__(
        cls, /, target_callable: _Optional[_Callable[..., _Any]] = None, **kwargs
    ):
        """
        This method is called when :class:`.DataclassedPartial` is subclassed.  It
        sets up functionality for configuring the callable.

        Args:
            target_callable (:obj:`callable`):  The function or class that the new
                :class:`.DataclassedPartial` will configure.
        """
        if target_callable is None:
            return
        if not callable(target_callable):
            raise TypeError(f"target_callable {target_callable} must be callable")

        cls._CALLABLE = staticmethod(target_callable)
        cls._ARG_SPEC = cls._get_target_callable_arg_spec(target_callable)
        cls._ARGS_DATACLASS = cls._make_dataclass_for_arg_spec(
            cls._ARG_SPEC, name=cls.__name__ + "Args"
        )
        cls._attach_arg_forwarding_properties()
        cls.__signature__ = _inspect.signature(target_callable)

    @staticmethod
    def _get_target_callable_arg_spec(
        target_callable: _Callable,
    ) -> _inspect.FullArgSpec:
        """
        Returns an argspec for ``target_callable`` but removes the self/cls arguments
        from the positional arguments.
        """
        spec = _inspect.getfullargspec(target_callable)
        args = spec.args
        defaults = spec.defaults
        if _inspect.isclass(target_callable) or _inspect.ismethod(target_callable):
            args = args[1:]  # exclude "self" from positional args
            if defaults is not None:
                defaults = defaults[-len(args) :]
        return _inspect.FullArgSpec(
            args=args,
            varargs=spec.varargs,
            varkw=spec.varkw,
            defaults=defaults,
            kwonlyargs=spec.kwonlyargs,
            kwonlydefaults=spec.kwonlydefaults,
            annotations=spec.annotations,
        )

    @staticmethod
    def _make_dataclass_for_arg_spec(
        arg_spec: _inspect.FullArgSpec, *, name: str
    ) -> _typing.Type:
        """
        Creates a dataclass type for holding a function's arguments.

        Args:
            arg_spec (:obj:`inspect.FullArgSpec`): The arg spec for the function.
            name (:obj:`str`): The name for the dataclass type.

        Returns:
            The new dataclass type with fields corresponding to the args, varargs,
            kwonlyargs, and kwargs paramter names from the argspec.  The defaults for
            each field also match the defaults from ``arg_spec``.
        """
        defaults = {}
        if arg_spec.defaults is not None:
            matching_args = arg_spec.args[-len(arg_spec.defaults) :]
            defaults.update(zip(matching_args, arg_spec.defaults))
        if arg_spec.kwonlydefaults is not None:
            defaults.update(arg_spec.kwonlydefaults)

        fields = []
        for arg in arg_spec.args + arg_spec.kwonlyargs:
            arg_type = arg_spec.annotations.get(arg, _typing.Any)
            default = defaults.get(arg, None)
            field = (arg, arg_type, default)
            fields.append(field)

        if arg_spec.varargs is not None:
            default = _dataclasses.field(default_factory=tuple)
            field = (arg_spec.varargs, _typing.Tuple[_typing.Any], default)
            fields.append(field)

        if arg_spec.varkw is not None:
            default = _dataclasses.field(default_factory=dict)
            field = (arg_spec.varkw, _typing.Dict[str, _typing.Any], default)
            fields.append(field)

        return _dataclasses.make_dataclass(name, fields)

    @classmethod
    def _attach_arg_forwarding_properties(cls):
        """
        Attaches properties that forward arg values from the configured_args dataclass
        to the parent DataclassedPartial object if there is no name conflict
        """
        for name in cls._ARGS_DATACLASS.__dataclass_fields__:
            if name in dir(cls):
                continue  # name conflict
            forwarder = _ConfiguredArgForwarder(name)
            prop = property(forwarder.get, forwarder.set)
            prop.__doc__ = f"The ``{name}`` configured argument."
            setattr(cls, name, prop)

    def __init__(self, *args, **kwargs):
        if self._CALLABLE is None:
            cls = self.__class__
            raise RuntimeError(
                f"Cannot init {cls} because it does not have a target callable."
            )
        self._configured_args = self._convert_callable_args_to_dataclass(
            *args, **kwargs
        )
        self._post_hooks = PartialPostHooks()
        self._context_hooks = PartialContextHooks()
        self._extended_context_hooks = PartialContextHooks()

    @property
    def post_hooks(self) -> "PartialPostHooks":
        """
        A :obj:`PartialPostHooks` instanec for modifying the result of
        :meth:`.__call__`.
        """
        return self._post_hooks

    @property
    def context_hooks(self) -> "PartialContextHooks":
        """
        A :obj:`PartialContextHooks` instance for modifying the context of
        :meth:`.__call__`. Note that ``post_hooks`` are *NOT* covered by this context.
        To include ``post_hooks`` in context, use ``extended_context_hooks`` instead.
        """
        return self._context_hooks

    @property
    def extended_context_hooks(self) -> "PartialContextHooks":
        """
        A :obj:`PartialContextHooks` instance for modifying the context of
        :meth:`.__call__`, including ``post_hooks``.
        """
        return self._extended_context_hooks

    @classmethod
    def create_subclass(
        cls,
        target_callable: _Callable[..., _Any],
        *,
        name: str,
        module_path: _Optional[str] = None,
    ) -> _typing.Type["DataclassedPartial"]:
        """
        Creates a new :class:`DataclassedPartial` subclass for a target callable.  The
        purpose of this type is to help configure arguments for the callable prior to
        calling it.

        Args:
            target_callable (:obj:`callable`): A class or other callable.
            name (:obj:`str`): The name of the new type.
            module_path (:obj:`str`, None): Optional module path of the target_callable
                to be registered.

        Returns:
            The new type.
        """
        return type(
            name,
            (cls,),
            {
                "__module__": target_callable.__module__
                if module_path is None
                else module_path
            },
            target_callable=target_callable,
        )

    @property
    def configured_args(self):
        """
        The configured arguments as a dataclass.  The fields of the dataclass match the
        signature of the target callable.

        .. tip::
            Except in cases of name conflicts, we can also access attributes of
            :attr:`configured_args` directly on the :obj:`.DataclassedPartial`.  For
            example, if an arg is named ``dim``, then ``config.dim`` forwards to
            ``config.configured_args.dim``.
        """
        return self._configured_args

    def update_configured_args(self, *new_args, **new_kwargs):
        """
        Updates :attr:`.configured_args` with values from arguments to the target
        callable.  For fields not specified by ``*new_args, **new_kwargs``, the
        value in :attr:`.configured_args` does not change.

        Args:
            new_args: Optional positional arguments to overwrite args specified in
                :attr:`configured_args`.  These args replace the first
                ``len(new_args)`` positional args (and *all* varargs if
                ``new_args`` contains varargs).
            new_kwargs: Optional keyword arguments.  These arguments replace additional
                named arguments not replaced by ``new_args``.
        """

        args, kwargs = self._merge_override_args_with_configured_args(
            *new_args, **new_kwargs
        )
        self._configured_args = self._convert_callable_args_to_dataclass(
            *args, **kwargs
        )

    def _convert_callable_args_to_dataclass(self, *args, **kwargs):
        """
        Given *args, **kwargs for the target callable, this helper creates a
        ``cls._ARGS_DATACLASS`` dataclass instance that holds these arguments.  The
        signature for dataclass creation is different from that of the target callable,
        and this conversion requires mapping everything in *args, **kwargs to the right
        parameter name of the callable.
        """

        # First map args to positional params (excluding varargs):
        result = self._ARGS_DATACLASS(
            **dict(zip(self._ARG_SPEC.args, args))
            # note: args may only cover some of ARG_SPEC.args (zip stops once any
            # sequence is exhausted). kwargs may cover additional positional args, and
            # any remaining entries in ARG_SPEC.args receive None as default
        )

        # Fill any positional args passed with keywords in **kwargs:
        for idx, name in enumerate(self._ARG_SPEC.args):
            if name not in kwargs:
                continue
            if idx < len(args):
                raise TypeError(f"got multiple values for argument '{name}'")
            setattr(result, name, kwargs.pop(name))

        if self._ARG_SPEC.varargs is not None:
            # Assign any remaining args to varargs:
            setattr(result, self._ARG_SPEC.varargs, args[len(self._ARG_SPEC.args) :])
        elif len(args) > len(self._ARG_SPEC.args):
            raise TypeError(
                f"{self._CALLABLE}() takes {len(self._ARG_SPEC.args)} positional "
                f"arguments but {len(args)} were given"
            )

        # Assign keyword-only args:
        for name in self._ARG_SPEC.kwonlyargs:
            if name in kwargs:
                setattr(result, name, kwargs.pop(name))

        if self._ARG_SPEC.varkw is not None:
            # Assign any remaining kwargs to varkw:
            setattr(result, self._ARG_SPEC.varkw, kwargs)
        elif len(kwargs) > 0:
            raise TypeError(f"Received unexpected keyword arguments {list(kwargs)}")

        return result

    def __call__(self, *override_args, **override_kwargs) -> _Any:
        """
        Calls the target callable with the configured args.

        Args:
            override_args: Optional positional arguments to override args specified in
                :attr:`configured_args`.  These args replace the first
                ``len(override_args)``
                positional args (and *all* varargs if ``override_args`` contains
                varargs).
            override_kwargs: Optional keyword override arguments.  These arguments
                replace any additional named arguments not overriden by
                ``override_args``.

        Calling a :class:`DataclassPartial` roughly equals to

        .. code-block:: python

            with context_hooks(): # enters a composite context defined by context hooks
                results = self.callable(*args, **kwargs)
                return apply_post_hooks(results)

        Returns:
            The result of calling the target callable with _optional_ override arguments
            under _optional_ contexts and application of _optional_ post hook.
        """

        args, kwargs = self._merge_override_args_with_configured_args(
            *override_args, **override_kwargs
        )

        _logger.debug(
            "Calling %s with args=%s, kwargs=%s", self._CALLABLE, args, kwargs
        )

        with _contextlib.ExitStack() as extended_stack:
            self.extended_context_hooks.enter_contexts(extended_stack)

            with _contextlib.ExitStack() as stack:
                self.context_hooks.enter_contexts(stack)
                result = self._CALLABLE(*args, **kwargs)  # pylint: disable=not-callable

            return self.post_hooks.apply(result)

    def _merge_override_args_with_configured_args(
        self, *override_args, **override_kwargs
    ):
        """
        Converts the :attr:`confiugred_args` dataclass back to *args, **kwargs for
        calling the target callable.  Then updates these *args, **kwargs using values
        from *override_args, **override_kwargs, returning the updated *args, **kwargs.
        """

        args, kwargs = self._convert_dataclass_to_callable_args(self.configured_args)

        # Replace positional arguments (and possibly varargs) with override_args:
        if len(override_args) <= len(self._ARG_SPEC.args):
            args[: len(override_args)] = override_args
        elif self._ARG_SPEC.varargs is not None:
            args = override_args
        else:
            raise TypeError(
                f"Callable takes {len(self._ARG_SPEC.args)} positional arguments "
                f"but {len(override_args)} positional override arguments were given"
            )

        # Replace positional arguments with override_kwargs:
        for idx, name in enumerate(self._ARG_SPEC.args):
            if name not in override_kwargs:
                continue
            if idx < len(override_args):
                # override_args already replaced the arg for name
                raise TypeError(
                    f"Received multiple override values for argument '{name}'"
                )
            args[idx] = override_kwargs.pop(name)

        # Replace keyword-only arguments with override_kwargs:
        kwargs.update(override_kwargs)

        return args, kwargs

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return vars(self) == vars(other)
        return False

    @classmethod
    def _convert_dataclass_to_callable_args(cls, configured_args):
        """
        Given a ``configured_args`` dataclass of type ``cls._ARGS_DATACLASS``, this
        method maps the dataclass back to *args, **kwargs for calling the target
        callable.
        """
        args = [getattr(configured_args, name) for name in cls._ARG_SPEC.args]
        if cls._ARG_SPEC.varargs is not None:
            args.extend(getattr(configured_args, cls._ARG_SPEC.varargs))
        kwargs = {
            name: getattr(configured_args, name) for name in cls._ARG_SPEC.kwonlyargs
        }
        if cls._ARG_SPEC.varkw is not None:
            kwargs.update(getattr(configured_args, cls._ARG_SPEC.varkw))
        return args, kwargs

    def __repr__(self):
        configured_args = _helpers.dataclass_to_dict(self.configured_args)
        items = list(configured_args.items())
        if len(self.post_hooks) > 0:
            items.append(("post_hooks", self.post_hooks))
        if len(self.context_hooks) > 0:
            items.append(("context_hooks", self.context_hooks))
        if len(self.extended_context_hooks) > 0:
            items.append(("extended_context_hooks", self.extended_context_hooks))
        reprs = [f"{key}={repr(value)}," for key, value in items]
        joined = "\n".join(reprs)
        joined = _textwrap.indent(joined, prefix=" " * 4)
        return f"{self.__class__.__name__}(\n{joined}\n)"


class _ConfiguredArgForwarder:
    """
    Helper class for defining properties on a DataclassedPartial that forward argument
    names to the ``configured_args`` dataclass.
    """

    def __init__(self, arg_name: str):
        self.arg_name = arg_name

    def get(self, partial: DataclassedPartial) -> _Any:
        return getattr(partial.configured_args, self.arg_name)

    def set(self, partial: DataclassedPartial, value: _Any):
        return setattr(partial.configured_args, self.arg_name, value)


class _PartialHooksBase:
    """A base class for a collection of hooks."""

    def __init__(self):
        self._hooks = {}
        self._next_hook_id = 0

    def register(self, hook: _Callable[[_Any], _Any]) -> "PartialHookHandle":
        """
        Registers a new hook and returns a :obj:`PartialHookHandle` for optionally
        removing it later.
        """
        hook_id = self._next_hook_id
        self._hooks[hook_id] = hook
        self._next_hook_id += 1
        return PartialHookHandle(hooks=self, hook_id=hook_id)

    def __iter__(self):
        return iter(self._hooks.values())

    def __len__(self) -> int:
        return len(self._hooks)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return list(self)[idx]
        return _helpers.get_item_by_index(self, idx)

    def __setitem__(self, idx, value):
        if isinstance(idx, slice):
            selected_keys = list(self._hooks)[idx]
            if len(selected_keys) != len(value):
                raise ValueError(
                    f"slice specifies {len(selected_keys)} layers but value has length "
                    f"{len(value)}"
                )
            for key, hook in zip(selected_keys, value):
                self._hooks[key] = hook
        else:
            key = _helpers.get_item_by_index(self._hooks, idx)
            self._hooks[key] = value

    def __delitem__(self, idx):
        if isinstance(idx, slice):
            selected_keys = list(self._hooks)[idx]
            for key in selected_keys:
                self._hooks.pop(key)
        else:
            key = _helpers.get_item_by_index(self._hooks, idx)
            self._hooks.pop(key)

    def __repr__(self):
        items = [repr(el) + "," for el in self]
        joined = "\n".join(items)
        joined = _textwrap.indent(joined, prefix=" " * 4)
        return f"{self.__class__.__name__}(\n{joined}\n)"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self._hooks == other._hooks
                and self._next_hook_id == other._next_hook_id
            )
        return False


class PartialPostHooks(_PartialHooksBase):
    """
    A collection of callbacks for modifying the result of
    :meth:`.DataclassedPartial.__call__`.  Each hook must take the return value of
    the :obj:`.DataclassedPartial` and return either a new return value or ``None``.
    """

    def apply(self, result: _Any) -> _Any:
        """
        Calls the hooks by passing ``result`` to each hook and updating the result with
        the returned value if it is not ``None``.  Calls hooks in the order of hook
        registration.
        """
        for hook in self:
            new_result = hook(result)
            if new_result is not None:
                result = new_result
        return result


class PartialContextHooks(_PartialHooksBase):
    """
    A collection of hooks for altering the context of
    :meth:`.DataclassedPartial.__call__`.  Each hook must take no arguments and return a
    context manager.
    """

    def enter_contexts(self, stack: _contextlib.ExitStack) -> None:
        """
        Calls each hook and enters the returned context.  This happens in the order of
        hook registration.

        Args:
            stack (:obj:`contextlib.ExitStack`): An exit stack for combining multiple
                context managers.
        """
        for hook in self:
            stack.enter_context(hook())


class PartialHookHandle:
    """
    A handle for removing a hook from a :obj:`.PartialPostHooks`.
    """

    def __init__(self, *, hooks: PartialPostHooks, hook_id: int):
        self._hooks_ref = _weakref.ref(hooks)
        self._hook_id = hook_id

    @property
    def _hooks(self):
        return self._hooks_ref()

    def remove(self) -> None:
        """
        Removes the registered hook.  Raises a :obj:`KeyError` if the hook has already
        been removed.
        """
        if self._hooks is None:
            return
        self._hooks._hooks.pop(self._hook_id)  # pylint: disable=protected-access

    def __enter__(self) -> "PartialHookHandle":
        return self

    def __exit__(self, type, value, tb):  # pylint: disable=redefined-builtin
        self.remove()



import abc
import tempfile
from pathlib import Path
from typing import Any, Optional, Union

import tamm.utils.json as tamm_json
from tamm.model_repo.publishing.config import PublishedModelConfig
from tamm.models.common import ModelConfig


class ModelPublisher(abc.ABC):
    def publish(
        self,
        payload: Union[Path, ModelConfig, PublishedModelConfig],
        model_id: Optional[str] = None,
        project_id: Optional[str] = None,
        version_id: Optional[str] = None,
    ):
        """
        Publishes one of the following to a model repository

        1) A :class:`PublishedModelConfig`
        2) A :class:`ModelConfig`
        3) A model config file path in :class:`pathlib.Path` or :obj:`str`
        4) Contents of ``directory`` (:class:`Path`, or :obj:`str`) which includes a
           config file (*.json) and ckpts

        Args:
            payload: :obj:`str`, :class:`Path`, :class:`ModelConfig` or
                     :class:`PublishedModelConfig`
            model_id: Model ID override name of the model
                          (required if ``payload`` is not :class:`PublishedModelConfig`)
            project_id: Optional project ID for supported backends
            version_id: Optional version ID for supported backends

        Returns: A handle of the published model


        """
        if isinstance(payload, PublishedModelConfig):
            return self.publish_published_model_config(payload)
        if isinstance(payload, ModelConfig):
            return self.publish_model_config(payload, model_id, project_id, version_id)
        if isinstance(payload, (str, Path)) and Path(payload).is_dir():
            return self.publish_directory(
                Path(payload), model_id, project_id, version_id
            )
        if isinstance(payload, (str, Path)) and Path(payload).is_file():
            return self.publish_file(Path(payload), model_id, project_id, version_id)

        raise ValueError(
            f"Cannot publish '{payload}' because it is not a recognized object. "
            f"Currently supports "
            f"1) PublishedModelConfig, "
            f"2) ModelConfig, "
            f"3) A path to config.json, and "
            f"4) A directory path containing config.json and checkpoints"
        )

    def publish_published_model_config(
        self,
        published_model_config: "PublishedModelConfig",
        model_id: Optional[str] = None,
        project_id: Optional[str] = None,
        version_id: Optional[str] = None,
    ) -> Any:
        published_model_config = self._maybe_update_model_id(
            published_model_config, model_id, project_id, version_id
        )
        if not bool(published_model_config.model_id):
            raise ValueError(
                f"published_model_config.model_id must be a valid string, "
                f"got {published_model_config.model_id}"
            )
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)
            tmp_published_config = tmp_path / "config.json"
            with open(tmp_published_config, "w", encoding="utf-8") as f:
                tamm_json.dump(published_model_config, f)
            return self.publish_directory(tmp_path)

    def _maybe_update_model_id(
        self,
        config: "PublishedModelConfig",
        model_id: Optional[str] = None,
        project_id: Optional[str] = None,
        version_id: Optional[str] = None,
    ) -> "PublishedModelConfig":
        resolved_model_id = config.model_id if model_id is None else model_id
        try:
            config.model_id = self._get_model_id(
                resolved_model_id, project_id, version_id
            )
        except ValueError:
            pass
        return config

    # pylint: disable=unused-argument
    def _get_model_id(
        self,
        base_model_id: Optional[str] = None,
        project_id: Optional[str] = None,
        version_id: Optional[str] = None,
    ) -> str:
        """
        Get model ID by parts
        Args:
            base_model_id: Basename of the model
            project_id: Optional project ID for supported backends
            version_id: Optional version ID for supported backends

        Returns: A valid model ID for the backend

        """
        if not base_model_id:
            raise ValueError(
                f"base_model_id must be a valid string, got {base_model_id}"
            )
        return base_model_id

    def publish_model_config(
        self,
        model_config: "ModelConfig",
        model_id: Optional[str] = None,
        project_id: Optional[str] = None,
        version_id: Optional[str] = None,
    ) -> Any:
        if model_id is None:
            raise ValueError(
                "`model_id` must be specified to publish subclasses of `ModelConfig`"
            )

        published_model_config = PublishedModelConfig(
            model_id=self._get_model_id(model_id, project_id, version_id),
            model_config=model_config,
        )
        return self.publish_published_model_config(published_model_config)

    def publish_file(
        self,
        file: Path,
        model_id: Optional[str] = None,
        project_id: Optional[str] = None,
        version_id: Optional[str] = None,
    ) -> Any:
        with open(file, "r", encoding="utf-8") as f:
            model_config = tamm_json.load(f)
        if isinstance(model_config, PublishedModelConfig):
            return self.publish_published_model_config(
                model_config,
                model_id=model_id,
                project_id=project_id,
                version_id=version_id,
            )
        if isinstance(model_config, ModelConfig):
            return self.publish_model_config(
                model_config,
                model_id=model_id,
                project_id=project_id,
                version_id=version_id,
            )
        raise ValueError(
            f"'{file}' must be either `PublishedModelConfig` or `ModelConfig`, "
            f"got {type(model_config)}"
        )

    def publish_directory(
        self,
        directory: Path,
        model_id: Optional[str] = None,
        project_id: Optional[str] = None,
        version_id: Optional[str] = None,
    ) -> Any:
        """
        Publishes the contents of ``directory`` as ``model_id``

        .. admonition:: Input requirements

            ``directory`` must have single ``config.json`` which is the valid
            :class:`PublishedModelConfig` for the candidate model.

        If implementation supports, checkpoints under ``directory`` will
        also be published.

        Args:
            directory: Directory to be published
            model_id: Model ID override name of the model
                          (required if ``payload`` is not :class:`PublishedModelConfig`)
            project_id: Optional project ID for supported backends
            version_id: Optional version ID for supported backends

        Returns:

        """
        if not directory.is_dir():
            raise ValueError(f"{directory} must be a directory")
        config_file = directory / "config.json"
        if not config_file.is_file():
            raise ValueError(
                f"PublishedModelConfig '{config_file}' must exist in this exact path"
            )
        with open(config_file, "r", encoding="utf-8") as f:
            model_config = tamm_json.load(f)
        if not isinstance(model_config, PublishedModelConfig):
            raise ValueError(
                f"'{model_config}' must be `PublishedModelConfig`, "
                f"got {type(model_config)}"
            )
        model_config = self._maybe_update_model_id(
            model_config, model_id, project_id, version_id
        )
        with open(config_file, "w", encoding="utf-8") as f:
            tamm_json.dump(model_config, f)
        return self._publish_directory(directory)

    @abc.abstractmethod
    def _publish_directory(
        self,
        directory: Path,
    ) -> Any:
        """
        Abstract method to be implemented in subclasses. Subclasses can safely
        assume there's a :class:`PublishedModelConfig` at ``<directory>/config.json``

        Args:
            directory: Directory which contains model artifacts (model config and ckpt)

        Returns: A handle of the published model

        """



"""
transformer.stack
=================

.. autoclass:: tamm.layers.TransformerStack
    :show-inheritance:
    :members:

.. autoclass:: tamm.layers.transformer.stack.TransformerStackOutput
    :members:

.. autoclass:: tamm.layers.ExtendedTransformerStack
    :show-inheritance:
    :members:

.. autoclass:: tamm.layers.transformer.stack.ExtendedTransformerStackOutput
    :members:
    :inherited-members:

.. autoclass:: tamm.layers.transformer.stack.ExtendedTransformerStackMode
    :members:
"""

import enum as _enum
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm import _helpers
from tamm._helpers import case_insensitive_lookup
from tamm.layers import functional as _tamm_F
from tamm.layers import side_outputs as _side_outputs
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.transformer import attention_mask as _attention_mask
from tamm.layers.transformer import kv_cache as _kv_cache
from tamm.layers.transformer import token_metadata as _token_metadata
from tamm.typing import ModuleBuilder as _ModuleBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder
from tamm.typing import OptionalTensor as _OptionalTensor
from tamm.utils import torch_utils as _torch_utils


@_torch_utils.torch_exportable_dataclass
class TransformerStackOutput:
    """A dataclass for holding outputs from a :class:`TransformerStack`."""

    last_hidden_state: _OptionalTensor = None
    """The (possibly normalized) output of the transformer stack."""

    hidden_states: _Optional[_List[_torch.Tensor]] = None
    """
    An optional list of ``n + 1`` hidden states, where ``n`` is the number of
    transformer layers.  This includes the (possibly normalized) input embeddings and
    the output from each layer.  This is ``None`` unless the layer stack produces side
    outputs with ``hidden_states``.
    """

    attentions: _Optional[_List[_torch.Tensor]] = None
    """
    An optional list of attention probabilities, one from each attention layer in the
    stack.  This is ``None`` unless the layer stack produces side outputs with
    ``attentions``.
    """

    expert_assignments: _Optional[_List[_torch.Tensor]] = None
    """
    An optional list of expert assignments, one from each Mixture of Expert layer in the
    stack.  This is ``None`` unless the layer stack produces side outputs with
    ``expert_assignments``.
    """


class TransformerStack(_nn.Module, _LayerMixin):
    """
    A generic transformer stack for encoders and decoders.  The output is the result of
    the following layer sequence:

    * An optional token type encoding
    * An optional positional encoding
    * An optional input transform
    * An optional input norm
    * A sequence of transformer layers
    * An optional output norm

    In addition, this layer uses attention mask and positional encoding layers to control
    the behavior of attention.

    This layer also has a :obj:`token_metadata_logic` child layer, which controls the
    value of token positions and segment ids.  Users can implement custom behavior by
    subclassing :class:`tamm.layers.transformer.token_metadata.TokenMetadataLogic`.

    Args:
        token_metadata_logic: An optional builder for a :class:`.TokenMetadataLogic`
            layer.  If ``None``, this defaults to :class:`.VanillaTokenMetadataLogic`.
        token_type_encoding:  An optional layer that transforms input embeddings as a
            function of the ``token_types`` passed to :meth:`.forward`.
        attention_mask: An optional attention mask builder for controlling which tokens
            attend to one another.
        positional_encoding: An optional position encoder builder for incorporating
            positions into the attention computation.
        secondary_positional_encodings: An optional
            :class:`.SecondaryPositionalEncodings` layer for passing additional
            positional encoding outputs to ``layers``.  This is useful for models that
            mix positional encoding types (such as local and global attention).
        input_transform: An optional builder for mapping target embeddings into
            different dimensions.
        input_norm: An optional norm builder for normalizing target embeddings.
        layers: A builder for the sequence of transformer layers.
        output_norm: An optional norm builder for normalizing outputs.
        embeddings_cast_dtype (:obj:`torch.dtype` or ``str``, optional): A target dtype
            for casting input embeddings, which should match the dtype for computation.
            If ``"auto"``, the layer attempts to infer the correct dtype.  If ``None``,
            the layer performs no cast.
    """

    def __init__(
        self,
        *,
        token_metadata_logic: _OptionalModuleOrBuilder = None,
        token_type_encoding: _OptionalModuleOrBuilder = None,
        attention_mask: _OptionalModuleOrBuilder = None,
        positional_encoding: _OptionalModuleOrBuilder = None,
        secondary_positional_encodings: _OptionalModuleOrBuilder = None,
        input_transform: _OptionalModuleOrBuilder = None,
        input_norm: _OptionalModuleOrBuilder = None,
        layers: _ModuleBuilder,
        output_norm: _OptionalModuleOrBuilder = None,
        embeddings_cast_dtype: _Optional[_Union[str, _torch.dtype]] = "auto",
    ):
        super().__init__()

        if token_metadata_logic is None:
            token_metadata_logic = _token_metadata.VanillaTokenMetadataLogic.Builder()
        if attention_mask is None:
            attention_mask = _attention_mask.AttentionMask.Builder()

        _helpers.append_children(
            # pylint: disable=duplicate-code
            self,
            token_metadata_logic=token_metadata_logic,
            token_type_encoding=token_type_encoding,
            attention_mask=attention_mask,
            positional_encoding=positional_encoding,
            secondary_positional_encodings=secondary_positional_encodings,
            input_transform=input_transform,
            input_norm=input_norm,
            layers=layers,
            output_norm=output_norm,
        )
        self.embeddings_cast_dtype = embeddings_cast_dtype

    @property
    def num_layers(self) -> int:
        """The number of transformer layers in the stack."""
        return self.layers.num_layers

    def forward(
        self,
        inputs: _torch.Tensor,
        *,
        segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
        attention_mask: _OptionalTensor = None,
        cross_attention_source: _OptionalTensor = None,
        cross_attention_mask: _OptionalTensor = None,
        aux_token_metadata: _Optional[_Dict[str, _Any]] = None,
    ) -> TransformerStackOutput:
        """
        Args:
            inputs (:obj:`torch.Tensor` or :obj:`dict`): Input target embeddings,
                typically with shape ``(batch_size, seq_len, hidden_dim)``.  The inputs
                can also have multiple sequence dimensions, in which case the layer
                flattens the input sequence and unflattens the output sequence.  The
                inputs can even be a :obj:`dict` the maps strings to embeddings, in
                which case the layer concatenates the inputs and splits the outputs.
            segment_ids (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that controls the attention mask for
                self attention layers.  For sequence ``i``, if
                ``segment_ids[i, j] != segment_ids[i, k]``, then tokens ``j`` and
                ``k`` do not attend to each other.  If unspecified, ``segment_ids``
                defaults to all ones.
            positions (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies the position of each
                token.  If unspecified, ``positions`` defaults to
                ``0, 1, 2, ...`` for each segment of each sequence.
            token_types (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies a "type" for each token.
                The ``token_type_encoding`` and ``attention_mask`` child layers may use
                this type information in different ways, depending on the model.
            kv_cache (:obj:`BaseKVCacheView`, optional): A KV cache view with cached
                keys and values for self-attention layers.
            attention_mask (:obj:`torch.Tensor`, optional): A boolean tensor with shape
                ``(batch_size, seq_len, kv_seq_len)``.  A ``True`` value indicates that
                the corresponding pair of tokens should take part in attention.  The
                ``kv_seq_len`` should be the same as ``seq_len`` unless using a
                ``kv_cache``, in which case it is the cache length.
            cross_attention_source (:obj:`torch.Tensor`, optional): Input source
                embeddings for cross attention layers with shape
                ``(batch_size, source_seq_len, hidden_dim)``.
            cross_attention_mask (:obj:`torch.Tensor`, optional): An attention mask for
                cross attention with shape ``(batch_size, seq_len, source_seq_len)``.
            aux_token_metadata (:obj:`dict`): A dictionary of optional auxiliary inputs for
                extending model behavior.  This gets attached to the ``token_metadata``
                input of several child layers, including the attention mask, layer
                sequence, and positional encoding layers.
        """

        # pylint: disable=too-many-locals

        x = _tamm_F.maybe_flatten_embeddings(inputs)
        segment_ids = _tamm_F.maybe_flatten_sequence(segment_ids)
        positions = _tamm_F.maybe_flatten_sequence(positions)
        token_types = _tamm_F.maybe_flatten_sequence(token_types)
        cross_attention_source = _tamm_F.maybe_flatten_embeddings(
            cross_attention_source
        )

        x = self._maybe_cast_inputs(x, kv_cache=kv_cache)
        cross_attention_source = self._maybe_cast_inputs(
            cross_attention_source, kv_cache=kv_cache
        )

        token_metadata = self.token_metadata_logic(
            embeddings=x,
            segment_ids=segment_ids,
            positions=positions,
            token_types=token_types,
            kv_cache=kv_cache,
            aux_metadata=aux_token_metadata,
        )

        if attention_mask is None:
            attention_side_inputs = self.attention_mask(
                token_metadata, compute_dtype=x.dtype
            )
        else:
            attention_side_inputs = {
                "attention_mask": attention_mask,
                "flash_attention_options": None,  # Flash does not support custom masks
            }

        if self.token_type_encoding is not None:
            x, attention_side_inputs = self.token_type_encoding(
                token_metadata=token_metadata,
                embeddings=x,
                attention_side_inputs=attention_side_inputs,
            )

        if self.secondary_positional_encodings is not None:
            secondary_attention_side_inputs = self.secondary_positional_encodings(
                token_metadata=token_metadata,
                embeddings=x,
                attention_side_inputs=attention_side_inputs,
            )
        else:
            secondary_attention_side_inputs = None

        if self.positional_encoding is not None:
            x, attention_side_inputs = self.positional_encoding(
                token_metadata=token_metadata,
                embeddings=x,
                attention_side_inputs=attention_side_inputs,
            )

        if self.input_transform is not None:
            x = self.input_transform(x)

        if self.input_norm is not None:
            x = self.input_norm(x)

        cross_attention_side_inputs = {
            "source": cross_attention_source,
            "attention_mask": cross_attention_mask,
        }

        x = self.layers(
            x,
            attention_side_inputs=attention_side_inputs,
            kv_cache=kv_cache,
            secondary_attention_side_inputs=secondary_attention_side_inputs,
            cross_attention_side_inputs=cross_attention_side_inputs,
            token_metadata=token_metadata,
        )

        if isinstance(x, _side_outputs.OutputWithSideOutputs):
            outputs = x.side_outputs
            x = x.output
        else:
            outputs = {}

        if self.output_norm is not None:
            x = self.output_norm(x)

        x = _tamm_F.maybe_unflatten_embeddings(x, original=inputs)

        outputs["last_hidden_state"] = x
        return TransformerStackOutput(**outputs)

    def _maybe_cast_inputs(
        self,
        tensor: _Union[_torch.Tensor, None],
        *,
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
    ) -> _Union[_torch.Tensor, None]:
        if tensor is None:
            return None
        if self.embeddings_cast_dtype is None:
            return tensor
        if self.embeddings_cast_dtype != "auto":
            return tensor.type(self.embeddings_cast_dtype)
        if kv_cache is not None and kv_cache.dtype is not None:
            dtype = kv_cache.dtype
        else:
            dtype = _helpers.get_dtype_after_autocast(tensor.dtype, tensor.device)
        if dtype is None:
            return tensor
        return tensor.type(dtype)


@_torch_utils.torch_exportable_dataclass
class ExtendedTransformerStackOutput(TransformerStackOutput):
    """A dataclass for holding outputs from an :class:`ExtendedTransformerStack`."""

    embeddings: _OptionalTensor = None
    """The output of the embedding layer."""

    predictions: _OptionalTensor = None
    """The output of the output transform layer."""

    loss: _OptionalTensor = None
    """The output of the loss layer."""


class ExtendedTransformerStackMode(str, _enum.Enum):
    """
    An :obj:`Enum` for controlling the behavior of
    :meth:`ExtendedTransformerStack.forward`.
    """

    #: Perform the full forward pass using all available layers.
    FULL = "FULL"

    #: Skip the embedding layer (user passes input embeddings).
    SKIP_EMBEDDING_LAYER = "SKIP_EMBEDDING_LAYER"

    #: Return the output of the embedding layer only.
    ONLY_EMBEDDING_LAYER = "ONLY_EMBEDDING_LAYER"

    #: Return the output of the transformer layers but skip the head layer.
    SKIP_OUTPUT_LAYER = "SKIP_OUTPUT_LAYER"

    #: Skip both the embedding and head layers.
    SKIP_EMBEDDING_AND_OUTPUT_LAYERS = "SKIP_EMBEDDING_AND_OUTPUT_LAYERS"

    @classmethod
    def _missing_(cls, value):
        return case_insensitive_lookup(cls, value)


class ExtendedTransformerStackOutputTransformBuildMode(str, _enum.Enum):
    """
    An :class:`Enum` for modifying the building of ``output_transform`` layers
    during :class:`ExtendedTransformerStack` initialization.  The main purpose is to
    support tying ``embedding`` and ``output_transform`` weights.
    """

    #: Pass no arguments to :meth:`builder.build` when building the output transform
    #: layer.
    BASIC = "BASIC"

    #: Pass the embedding layer as the only argument when building the layer.
    PASS_EMBEDDING = "PASS_EMBEDDING"

    @classmethod
    def _missing_(cls, value):
        return case_insensitive_lookup(cls, value)


class ExtendedTransformerStack(TransformerStack, _LayerMixin):
    """
    This layer extends :class:`.TransformerStack` by adding optional embedding, output
    transform, and loss layers.

    Args:
        embedding: An optional builder for a layer that transforms inputs into token
            embeddings.
        token_metadata_logic: An optional builder for a :class:`.TokenMetadataLogic` layer.
        token_type_encoding: An optional builder for a layer that encodes token types.
        attention_mask: An optional attention mask builder for controlling which tokens
            attend to one another.
        positional_encoding: An optional position encoder builder for incorporating
            positions into the attention computation.
        secondary_positional_encodings: An optional
            :class:`.SecondaryPositionalEncodings` layer for passing additional
            positional encoding outputs to ``layers``.
        input_transform: An optional builder for mapping target embeddings into
            different dimensions.
        input_norm: An optional norm builder for normalizing target embeddings.
        layers: A builder for the sequence of transformer layers.
        output_norm: An optional norm builder for normalizing outputs.
        output_transform: An optional builder for mapping (possibly normalized) outputs
            of the transformer layers into predictions.
        output_transform_build_mode (:obj:`str`, optional): An optional member of
            :class:`.ExtendedTransformerStackOutputTransformBuildMode` to modify the
            building of ``output_transform``.
        loss: An optional builder for computing a loss value from the predictions and
            labels.
        embeddings_cast_dtype (:obj:`torch.dtype` or ``str``, optional): A target dtype
            for casting input embeddings, which should match the dtype for computation.
            If ``"auto"``, the layer attempts to infer the correct dtype.  If ``None``,
            the layer performs no cast.
    """

    def __init__(
        self,
        *,
        embedding: _OptionalModuleOrBuilder = None,
        token_metadata_logic: _OptionalModuleOrBuilder = None,
        token_type_encoding: _OptionalModuleOrBuilder = None,
        attention_mask: _OptionalModuleOrBuilder = None,
        positional_encoding: _OptionalModuleOrBuilder = None,
        secondary_positional_encodings: _OptionalModuleOrBuilder = None,
        input_transform: _OptionalModuleOrBuilder = None,
        input_norm: _OptionalModuleOrBuilder = None,
        layers: _ModuleBuilder,
        output_norm: _OptionalModuleOrBuilder = None,
        output_transform: _OptionalModuleOrBuilder = None,
        output_transform_build_mode: _Optional[str] = "basic",
        loss: _OptionalModuleOrBuilder = None,
        embeddings_cast_dtype: _Optional[_Union[str, _torch.dtype]] = "auto",
    ):
        super().__init__(
            layers=layers,
            token_metadata_logic=token_metadata_logic,
            token_type_encoding=token_type_encoding,
            attention_mask=attention_mask,
            positional_encoding=positional_encoding,
            secondary_positional_encodings=secondary_positional_encodings,
            input_transform=input_transform,
            input_norm=input_norm,
            output_norm=output_norm,
            embeddings_cast_dtype=embeddings_cast_dtype,
        )
        _helpers.prepend_children(self, embedding=embedding)
        self._init_output_transform(output_transform, mode=output_transform_build_mode)
        _helpers.append_children(self, loss=loss)

    def _init_output_transform(self, output_transform, *, mode):
        mode = _helpers.get_enum_member_from_name(
            ExtendedTransformerStackOutputTransformBuildMode, mode
        )
        if mode is ExtendedTransformerStackOutputTransformBuildMode.PASS_EMBEDDING:
            args = (self.embedding,)
        else:
            args = ()
        self.output_transform = _helpers.maybe_build_module(output_transform, *args)

    @property
    def head_layer(self) -> _Union[_nn.Module, None]:
        """
        Returns the head layer of the model.  In this case, it's the same as
        ``output_transform``.
        """
        return self.output_transform

    def forward(
        self,
        inputs: _Any,
        *,
        labels: _Any = None,
        segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
        attention_mask: _OptionalTensor = None,
        cross_attention_source: _OptionalTensor = None,
        cross_attention_mask: _OptionalTensor = None,
        aux_token_metadata: _Optional[_Dict[str, _Any]] = None,
        mode: _Union[ExtendedTransformerStackMode, str] = "full",
    ) -> ExtendedTransformerStackOutput:
        """
        Args:
            inputs (:obj:`torch.Tensor`): Inputs to the ``embedding`` layer.  For some
                values of ``mode``, this should instead be input embeddings.
            labels: Optional labels for the ``loss`` layer.  The loss layer should
                accept two positional arguments: (1) the output of ``output_transform``
                and (2) ``labels``.
            segment_ids (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that controls the attention mask for
                self attention layers.  For sequence ``i``, if
                ``segment_ids[i, j] != segment_ids[i, k]``, then tokens ``j`` and
                ``k`` do not attend to each other.  If unspecified, ``segment_ids``
                defaults to all ones.
            positions (:obj:`torch.Tensor`, optional): Integer tensor with shape
                ``(batch_size, seq_len)`` that specifies the position of each
                token.  If unspecified, ``positions`` defaults to
                ``0, 1, 2, ...`` for each segment of each sequence.
            token_types (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies a "type" for each token.
                The ``token_type_encoding`` and ``attention_mask`` child layers may use
                this type information in different ways, depending on the model.
            kv_cache (:obj:`BaseKVCacheView`, optional): A KV cache view with cached
                keys and values for self-attention layers.
            attention_mask (:obj:`torch.Tensor`, optional): A boolean tensor with shape
                ``(batch_size, seq_len, kv_seq_len)``.  A ``True`` value indicates that
                the corresponding pair of tokens should take part in attention.  The
                ``kv_seq_len`` should be the same as ``seq_len`` unless using a
                ``kv_cache``, in which case it is the cache length.
            cross_attention_source (:obj:`torch.Tensor`, optional): Input source
                embeddings for cross attention layers with shape
                ``(batch_size, source_seq_len, hidden_dim)``.
            cross_attention_mask (:obj:`torch.Tensor`, optional): An attention mask for
                cross attention with shape ``(batch_size, seq_len, source_seq_len)``.
            aux_token_metadata (:obj:`dict`): A dictionary of optional auxiliary inputs for
                extending model behavior.  This gets attached to the ``token_metadata``
                input of several child layers, including the attention mask, layer
                sequence, and positional encoding layers.
            mode (:obj:`str` or :obj:`.ExtendedTransformerStackMode`, optional): A
                mode for controlling the forward behavior.  Defaults to ``FULL``.
        """

        mode = _helpers.get_enum_member_from_name(ExtendedTransformerStackMode, mode)

        if self._should_skip_embedding_layer(mode):
            x = inputs
        else:
            x = self.embedding(inputs)
        x, outputs = self._maybe_merge_side_outputs(x, {})
        outputs["embeddings"] = x

        if mode is ExtendedTransformerStackMode.ONLY_EMBEDDING_LAYER:
            return ExtendedTransformerStackOutput(**outputs)

        x = super().forward(
            inputs=x,
            segment_ids=segment_ids,
            positions=positions,
            token_types=token_types,
            kv_cache=kv_cache,
            attention_mask=attention_mask,
            cross_attention_source=cross_attention_source,
            cross_attention_mask=cross_attention_mask,
            aux_token_metadata=aux_token_metadata,
        )
        x, outputs = self._merge_transformer_stack_outputs(x, outputs)

        if self._should_skip_output_transform(mode):
            return ExtendedTransformerStackOutput(**outputs)

        x = self.output_transform(x)
        x, outputs = self._maybe_merge_side_outputs(x, outputs)
        outputs["predictions"] = x

        if self.loss is not None and labels is not None:
            outputs["loss"] = self.loss(x, labels)

        return ExtendedTransformerStackOutput(**outputs)

    @staticmethod
    def _maybe_merge_side_outputs(
        new_outputs: _Any, old_side_outputs: _Dict[str, _Any]
    ) -> _Tuple[_Any, _Dict[str, _Any]]:
        if isinstance(new_outputs, _side_outputs.OutputWithSideOutputs):
            new_side_outputs = _side_outputs.merge_side_outputs(
                old_side_outputs, new_outputs.side_outputs
            )
            new_outputs = new_outputs.output
        else:
            new_side_outputs = old_side_outputs

        return new_outputs, new_side_outputs

    @staticmethod
    def _merge_transformer_stack_outputs(
        new_outputs: TransformerStackOutput, old_side_outputs: _Dict[str, _Any]
    ) -> _Tuple[_Any, _Dict[str, _Any]]:
        new_side_outputs = dict(
            old_side_outputs.items()
            # this is a torch.compile-friendly shallow copy of the dict, since
            # torch.compile graph-breaks on old_side_outputs.copy()
        )
        new_side_outputs["last_hidden_state"] = new_outputs.last_hidden_state
        new_side_outputs["hidden_states"] = new_outputs.hidden_states
        new_side_outputs["attentions"] = new_outputs.attentions
        new_side_outputs["expert_assignments"] = new_outputs.expert_assignments
        return new_outputs.last_hidden_state, new_side_outputs

    def _should_skip_embedding_layer(self, mode: ExtendedTransformerStackMode) -> bool:
        if self.embedding is None:
            return True
        return mode in (
            ExtendedTransformerStackMode.SKIP_EMBEDDING_LAYER,
            ExtendedTransformerStackMode.SKIP_EMBEDDING_AND_OUTPUT_LAYERS,
        )

    def _should_skip_output_transform(self, mode: ExtendedTransformerStackMode) -> bool:
        if self.output_transform is None:
            return True
        return mode in (
            ExtendedTransformerStackMode.SKIP_OUTPUT_LAYER,
            ExtendedTransformerStackMode.SKIP_EMBEDDING_AND_OUTPUT_LAYERS,
        )



"""
Command for listing contents of tamm registries (such as the available models or
tokenizers).
"""

from tamm._cli._listers import get_registered_listers, get_registered_listers_names
from tamm._cli.common import argument, command

# pylint: disable=redefined-builtin


@command("ls", help_text="List a tamm object registry")
@argument("-w", "--wide", action="store_true", help="Disable output truncation")
@argument("-l", "--long", action="store_true", help="Include descriptions")
@argument(
    "-d", "--show-deprecated", action="store_true", help="Include deprecated items"
)
@argument("-a", "--all", action="store_true", help="Include all discoverable items")
@argument("registry", type=str, choices=get_registered_listers_names())
def list_subcommand(registry, all, long, wide, show_deprecated):
    get_registered_listers()[registry](all, long, wide, show_deprecated).print()



"""
tamm
----

Model API
^^^^^^^^^
.. autofunction:: tamm.create_model
.. autofunction:: tamm.create_adapted_model
.. autofunction:: tamm.create_model_builder
.. autofunction:: tamm.create_model_config
.. autofunction:: tamm.is_adapted_model_name
.. autofunction:: tamm.is_model_builder_name
.. autofunction:: tamm.is_model_config_name
.. autofunction:: tamm.is_model_name
.. autofunction:: tamm.list_adapted_models
.. autofunction:: tamm.list_model_builders
.. autofunction:: tamm.list_model_configs
.. autofunction:: tamm.list_models

Tokenizer API
^^^^^^^^^^^^^
See :ref:`tamm.tokenizers <tamm_tokenizers>`

Converter API
^^^^^^^^^^^^^
.. autofunction:: tamm.list_converters

Checkpoint API
^^^^^^^^^^^^^^
.. autofunction:: tamm.load
.. autofunction:: tamm.save


Logging
^^^^^^^

.. data:: tamm.logger

    The root |tamm| :obj:`logging.Logger`.


Runtime configuration
^^^^^^^^^^^^^^^^^^^^^

.. data:: tamm.rc

    The global :obj:`.RuntimeConfiguration` instance.


|tamm| Attributes
^^^^^^^^^^^^^^^^^
.. automodule:: tamm.context_vars

"""

# pylint: disable=wrong-import-position

# isort: off
# configure rc and logger first because they may be used during import
from tamm.runtime_configuration import rc  # pylint: disable=unused-import
from tamm import _logger

logger = _logger.init_logger()
# isort: on

from tamm import (
    _adapters_v1,
    _compat,
    _plugin,
    adapters,
    ao,
    converters,
    generation,
    layers,
    models,
    preprocessors,
    utils,
)
from tamm._version import __version__
from tamm.context_vars import (
    first_token_generation_context,
    freeze_params_flag_context,
    get_default_freeze_params_flag,
    get_default_pretrained_flag,
    get_first_token_generation_flag,
    pretrained_flag_context,
    resolve_device,
    resolve_freeze_params,
    resolve_pretrained_flag,
    set_default_freeze_params_flag,
    set_default_pretrained_flag,
)
from tamm.converters import list_converters, load, save
from tamm.model_repo import (
    create_adapted_model,
    create_model,
    create_model_builder,
    create_model_config,
    is_adapted_model_name,
    is_model_builder_name,
    is_model_config_name,
    is_model_name,
    list_adapted_models,
    list_model_builders,
    list_model_configs,
    list_models,
)
from tamm.tokenizers import (
    create_tokenizer,
    create_tokenizer_config,
    create_tokenizer_from_model_id,
    create_tokenizer_from_tokenizer_id,
    list_tokenizers,
)

_plugin.execute_core_import_callbacks()

_compat.execute_backward_compatibility_imports()

__all__ = [
    "logger",
    "adapters",
    "ao",
    "converters",
    "layers",
    "models",
    "preprocessors",
    "utils",
    "create_adapted_model",
    "create_model",
    "create_model_builder",
    "create_model_config",
    "generation",
    "is_adapted_model_name",
    "is_model_builder_name",
    "is_model_config_name",
    "is_model_name",
    "list_adapted_models",
    "list_model_builders",
    "list_model_configs",
    "list_models",
    "__version__",
    "list_converters",
    "load",
    "save",
    "rc",
    "create_tokenizer",
    "create_tokenizer_config",
    "create_tokenizer_from_model_id",
    "create_tokenizer_from_tokenizer_id",
    "list_tokenizers",
    "freeze_params_flag_context",
    "first_token_generation_context",
    "pretrained_flag_context",
    "get_default_freeze_params_flag",
    "get_first_token_generation_flag",
    "get_default_pretrained_flag",
    "set_default_freeze_params_flag",
    "set_default_pretrained_flag",
    "resolve_device",
    "resolve_pretrained_flag",
    "resolve_freeze_params",
]



"""
tamm.utils
----------
"""

from tamm.utils import (
    axlearn_utils,
    json,
    partial,
    transformers_utils,
    user_dir_utils,
    vision_utils,
)
from tamm.utils._torch_compatibility import _is_same_device_type
from tamm.utils.optional_bool import OptionalBool
from tamm.utils.registry import RegistrySpec

__all__ = [
    "axlearn_utils",
    "json",
    "partial",
    "transformers_utils",
    "user_dir_utils",
    "vision_utils",
    "OptionalBool",
    "RegistrySpec",
]



"""HF transformers model configuration for tamm CausalLMTransformer"""

import logging as _logging
from typing import Optional as _Optional

import tamm as _tamm
from tamm.hf.transformers import configuration_utils as _configuration_utils

_logger = _logging.getLogger(__name__)


class TammCausalLMConfig(_configuration_utils.TammPretrainedConfig):
    """
    Config for CausalLM models.

    Args:
        tamm_config (:obj:`.ModelConfig`): A |tamm| config for a
            :obj:`.CausalLMTransformer`.
        **kwargs: Keyword arguments for default fields in
            :obj:`transformers.PretrainedConfig`.
    """

    model_type = "tamm_causal_lm"

    def __init__(
        self, tamm_config: _Optional[_tamm.models.common.ModelConfig] = None, **kwargs
    ):
        kwargs.setdefault(
            "use_cache", True
        )  # required to pass use_cache to from_pretrained
        super().__init__(tamm_config=tamm_config, **kwargs)
        self.is_decoder = True
        self.is_encoder_decoder = False

    @property
    def tie_word_embeddings(self):
        try:
            model = self.tamm_config.create_model(device="meta")
        except Exception:
            return False
        return getattr(model, "is_embedding_tied_to_output_transform", False)

    @tie_word_embeddings.setter
    def tie_word_embeddings(self, value):  # pylint: disable=unused-argument
        _logger.debug("ignoring call to TammCausalLMConfig.tie_word_embeddings setter")

    @property
    def vocab_size(self) -> int:
        """
        The model's vocab size.  We include this property because some parts of the
        HuggingFace ecosystem require it (specifically beam search decoding as of
        ``transformers==4.50``).

        This attribute is not standardized in |tamm|, so it may not work for all
        models.  Please contact the |tamm| team if you encounter issues with this.
        """
        if self.tamm_config is None:
            raise RuntimeError("Cannot provide vocab_size, since tamm_config is None")
        if not hasattr(self.tamm_config, "vocab_size"):
            raise RuntimeError(
                "Cannot provide vocab_size, since tamm_config does not have an "
                "attribute named 'vocab_size'.  Please contact the tamm team "
                "to support this use case."
            )
        return self.tamm_config.vocab_size



"""Command for converting model checkpoints from a different format to tamm"""

import logging
from typing import Optional

import torch

from tamm import converters
from tamm._cli.common import argument, command

logger = logging.getLogger(__name__)


@command(
    "to-tamm", help_text="Convert model checkpoints from a different format to tamm"
)
@argument(
    "--dtype",
    type=str,
    help="Optional output dtype, such as float16, float32, etc",
)
@argument("dest", type=str, help="Destination path for the converted model state")
@argument(
    "source",
    type=str,
    help="Source path to the model state to convert to tamm",
)
@argument(
    "converter-id",
    type=str,
    help=(
        "A converter id that specifies how to do the conversion--use "
        "``tamm ls converters -l`` for descriptions of available converters"
    ),
)
def to_tamm(converter_id, source, dest, dtype: Optional[str] = None):
    state_dict = converters.load(source, converter=converter_id, dtype=dtype)
    logger.info("Saving tamm state dict to %s", dest)
    torch.save(state_dict, dest)
    logger.info("Done!")



"""
adapters.layer_adapters
=======================

Common
^^^^^^

.. autoclass:: tamm.adapters.LayerAdapter
    :show-inheritance:
    :members:

Compositions
^^^^^^^^^^^^

.. autoclass:: tamm.adapters.layer_adapters.CompositeInputTransform
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_adapters.CompositeOutputTransform
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_adapters.AveragedInputTransforms
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_adapters.AveragedOutputTransforms
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_adapters.StackedInputTransforms
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_adapters.StackedOutputTransforms
    :show-inheritance:
    :members:

LoRA
^^^^

.. autoclass:: tamm.adapters.layer_adapters.LoRA
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_adapters.BatchedLoRA
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_adapters.LoRAFusedMultiOutputLinear
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_adapters.BatchedLoRAFusedMultiOutputLinear
    :show-inheritance:
    :members:
"""

from tamm._adapters_v1.layer_adapters.average import (
    AveragedInputTransforms,
    AveragedOutputTransforms,
)
from tamm._adapters_v1.layer_adapters.common import (
    CompositeInputTransform,
    CompositeOutputTransform,
    LayerAdapter,
    LayerAdapterConfig,
)
from tamm._adapters_v1.layer_adapters.lora import (
    BatchedLoRA,
    BatchedLoRAFusedMultiOutputLinear,
    LoRA,
    LoRAFusedMultiOutputLinear,
)
from tamm._adapters_v1.layer_adapters.stack import (
    StackedInputTransforms,
    StackedOutputTransforms,
)



import json
from typing import Any, Dict, List

import tamm.utils.json as _tamm_json
from tamm.preprocessors.base import Preprocessor
from tamm.preprocessors.registry import PreprocessorRegistry
from tamm.utils.uri.uri_handler import _URIHandler, is_uri_or_posix


class URIHandlerPreprocessorRegistry(PreprocessorRegistry):
    def _create(self, identifier: str, *args, **kwargs) -> "Preprocessor":
        if not is_uri_or_posix(identifier):
            raise KeyError(
                f"{identifier} is neither a URL nor posix path, {self} cannot handle this"
            )

        with _URIHandler().open(identifier) as f:
            return _tamm_json.load(f)

    def list_objects(self, *args, **kwargs) -> List[str]:
        return []

    def describe(self, identifier: str, *args, **kwargs) -> Dict[str, Any]:
        if not is_uri_or_posix(identifier):
            raise KeyError(
                f"{identifier} is neither a URL nor posix path, {self} cannot handle this"
            )

        with _URIHandler().open(identifier) as f:
            return json.load(f)



import logging as _logging

from tamm import _helpers
from tamm.converters.models import common as _common
from tamm.converters.models import registry as _registry
from tamm.typing import OptionalDtypeOrString as _OptionalDtypeOrString

_logger = _logging.getLogger(__name__)


def convert_to_tamm_state_dict(state_dict, converter):
    """
    Converts a state dict from another format to tamm format.

    Args:
        state_dict (:obj:`dict`): A state dict to convert.
        converter (:obj:`str` or :obj:`.ModelStateDictConverter`, optional): A converter
            id or converter instance for converting ``obj`` to tamm format.

    Returns:
        The converted state dict in tamm format.
    """
    if not isinstance(converter, _common.ModelStateDictConverter):
        converter = _registry.create_converter_from_other_state_dict(
            converter, state_dict
        )
    _logger.info(
        "Converting state dict to tamm format with %s", converter.__class__.__name__
    )
    return converter.convert_to_tamm(state_dict)


def load(f, *, converter=None, dtype: _OptionalDtypeOrString = None, **kwargs):
    """
    Loads a model checkpoint and converts it to tamm format.

    Args:
        f (:obj:`str` or :obj:`PathLike`): The filepath to the model state to load.
        converter (:obj:`str` or :obj:`.ModelStateDictConverter`, optional): A converter
            id or converter instance for converting the loaded state to a tamm model
            state dict.  This value is ``None`` by default, in which case the function
            behaves like :func:`torch.load` .  Use :func:`tamm.list_converters` or
            ``tamm ls converters -l`` for a list of available converter ids.
        dtype (:obj:`torch.dtype` or :obj:`str`, optional): Output dtype, such as
            ``float16``, ``float32``, etc.  After converting the state to tamm format,
            the function casts the state's floating point tensors to this dtype.  The
            default is ``None``, which results in no cast.
        **kwargs: Keyword arguments to forward to the :meth:`load_other_state_dict`
            method of the converter.

    Returns:
        The converted state dict in tamm format.
    """
    if converter is None:
        return _helpers.safe_torch_load(f, **kwargs)

    if isinstance(converter, str):
        converter_cls = _registry.get_converter_cls(converter)
        _logger.info("Loading %s with converter %s", f, converter_cls.__name__)
        obj = converter_cls.load_other_state_dict(f, **kwargs)
        converter = converter_cls.from_other_state_dict(obj)
    else:
        _logger.info("Loading %s with converter %s", f, converter.__class__.__name__)
        obj = converter.load_other_state_dict(f, **kwargs)

    _logger.info("Converting state to tamm format")
    state = converter.convert_to_tamm(obj)
    _helpers.maybe_cast_state_dict_(state, dtype=dtype)
    return state



from typing import Any as _Any

import torch.nn as _nn

from tamm._adapters_v1.layer_adapters.common import AdapterMode as _AdapterMode
from tamm._adapters_v1.layer_adapters.common import (
    CompositeInputTransform as _CompositeInputTransform,
)
from tamm._adapters_v1.layer_adapters.common import (
    CompositeOutputTransform as _CompositeOutputTransform,
)
from tamm._adapters_v1.layer_adapters.common import (
    MergeableLayerAdapterMixin as _MergeableLayerAdapterMixin,
)


class StackedInputTransforms(_CompositeInputTransform, _MergeableLayerAdapterMixin):
    """
    A composite :class:`LayerAdapter` that stacks together input transformations from
    multiple child adapters.  The behavior is equivalent to the following:

    .. code-block:: python

        for adapter in adapters.values():
            args, kwargs = adapter(TRANSFORM_INPUTS, args=args, kwargs=kwargs)
        outputs = wrapped(args, kwargs)

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`LayerAdapter`): A dictionary
            that maps child adapter IDs to corresponding adapters.  Note that the
            ordering of the values in ``adapters`` determines the ordering of the stack.
    """

    def _transform_inputs(self, *, args, kwargs):
        for adapter in self._child_adapters.values():
            args, kwargs = adapter(
                mode=_AdapterMode.TRANSFORM_INPUTS, args=args, kwargs=kwargs
            )
        return args, kwargs

    def merge_adapter(self, wrapped_module: _nn.Module):
        if not all(
            getattr(adapter, "merge_adapter")
            for adapter in self._child_adapters.values()
        ):
            raise RuntimeError("All child adapters should be mergeable adapters.")

        for adapter in self._child_adapters.values():
            adapter.merge_adapter(wrapped_module)


class StackedOutputTransforms(_CompositeOutputTransform, _MergeableLayerAdapterMixin):
    """
    A composite :class:`LayerAdapter` that stacks together output transformations from
    multiple child adapters.  The behavior is equivalent to the following:

    .. code-block:: python

        outputs = wrapped(args, kwargs)
        for adapter in adapters.values():
            outputs = adapter(
                TRANSFORM_OUTPUTS,
                args=args,
                kwargs=kwargs,
                transformed_args=args,
                transformed_kwargs=kwargs,
                outputs=outputs,
            )

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`LayerAdapter`): A dictionary
            that maps child adapter IDs to corresponding adapters.  Note that the
            ordering of the values in ``adapters`` determines the ordering of the stack.
    """

    def _transform_outputs(
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _Any,
    ):
        for adapter in self._child_adapters.values():
            outputs = adapter(
                mode=_AdapterMode.TRANSFORM_OUTPUTS,
                outputs=outputs,
                args=args,
                kwargs=kwargs,
                transformed_args=transformed_args,
                transformed_kwargs=transformed_kwargs,
            )
        return outputs

    def merge_adapter(self, wrapped_module: _nn.Module):
        if not all(
            getattr(adapter, "merge_adapter")
            for adapter in self._child_adapters.values()
        ):
            raise RuntimeError("All child adapters should be mergeable adapters.")
        for adapter in self._child_adapters.values():
            adapter.merge_adapter(wrapped_module)



"""Command for printing tamm's version."""

from tamm import __version__
from tamm._cli.common import command


@command("version", help_text="Print tamm's version")
def print_version():
    print(__version__)



import enum as _enum
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm import _helpers
from tamm._helpers import case_insensitive_lookup
from tamm.ao.layers import functional as _tamm_ao_F
from tamm.layers import functional as _tamm_F


class FakeQuantizeObserverMode(str, _enum.Enum):
    """An enum for controlling observer usage in :obj:`FakeQuantize` layers."""

    ENABLED = "enabled"
    """
    Call the observer to update quantization params during very forward pass, even when
    in eval mode.
    """

    DISABLED = "disabled"
    """Never call the observer so that quantization params remain frozen."""

    ONLY_TRAINING = "only_training"
    """
    Only call the observer when the :obj:`FakeQuantize` layer is in training mode.
    """

    @classmethod
    def _missing_(cls, value):
        return case_insensitive_lookup(cls, value)


class FakeQuantize(_nn.Module):
    """
    A layer for applying fake quantization (quantization immediately followed by
    dequantization) to simulate the impact of quantization.  This layer is suitable for
    quantization-aware training, since gradients pass through the fake quantization op
    (i.e., the backward pass ignores the quantization).

    Args:
        observer (:obj:`nn.Module`): An observer layer, such as
            :obj:`.SimpleEMAMinMaxObserver` or
            :obj:`torch.ao.quantization.MinMaxObserver`.
            This layer must have a :meth:`calculate_qparams` method as well as
            :attr:`quant_min` and :attr:`quant_max` attributes.  Its forward pass should
            optionally update statistics for :meth:`calculate_qparams`.
        enable_observer (:obj:`FakeQuantizeObserverMode`, :obj:`str`, or :obj:`bool`):
            Option that controls when to call the observer to update quantization
            params.  Defaults to ``"only_training"``.  A ``True`` value resolves to
            ``"enabled"`` mode, and ``False`` resolves to ``"disabled"``.  Note that
            in dynamic mode, the forward pass always calls the observer, regardless of
            whether it is enabled.
        enable_fake_quant (:obj:`bool`, optional): A flag that when ``False`` bypasses
            the fake quantization during :meth:`forward`.  Defaults to ``True``.  This
            flag does not affect the behavior of ``enable_observer``.
        dynamic (:obj:`bool`): If True, quantization parameters are reset every
            iteration.
        cast_dtype (:obj:`torch.dtype`, optional): An optional dtype for casting the
            input tensor and scale prior to fake quantization.  Defaults to ``None``.
        reciprocal_mul (:obj:`bool`, optional): A flag to multiply the features by
            the reciprocal of the quantization scales instead of dividing.
    """

    def __init__(
        self,
        observer: _nn.Module,
        enable_observer: _Union[str, FakeQuantizeObserverMode, bool] = "only_training",
        enable_fake_quant: bool = True,
        dynamic=False,
        cast_dtype=None,
        reciprocal_mul=False,
    ):
        super().__init__()

        self.activation_post_process = observer
        self.observer_mode = self._resolve_observer_mode(enable_observer)
        self.is_fake_quant_enabled = enable_fake_quant
        self.is_dynamic = dynamic
        self.cast_dtype = cast_dtype
        self.use_reciprocal_mul = reciprocal_mul

    @property
    def is_observer_enabled(self) -> bool:
        if self.observer_mode is FakeQuantizeObserverMode.ENABLED:
            return True
        if self.observer_mode is FakeQuantizeObserverMode.DISABLED:
            return False
        if self.observer_mode is FakeQuantizeObserverMode.ONLY_TRAINING:
            return self.training
        raise ValueError(f"Observer mode {self.observer_mode} not supported")

    @is_observer_enabled.setter
    def is_observer_enabled(
        self, value: _Union[str, FakeQuantizeObserverMode, bool]
    ) -> None:
        self.observer_mode = self._resolve_observer_mode(value)

    @staticmethod
    def _resolve_observer_mode(
        value: _Union[str, FakeQuantizeObserverMode, bool]
    ) -> FakeQuantizeObserverMode:
        if value is True:
            return FakeQuantizeObserverMode.ENABLED
        if value is False:
            return FakeQuantizeObserverMode.DISABLED
        return _helpers.get_enum_member_from_name(FakeQuantizeObserverMode, value)

    @property
    def ch_axis(self):
        return getattr(self.activation_post_process, "ch_axis", None)

    def forward(self, tensor):  # pylint: disable=invalid-name
        """
        Calls the ``observer`` layer with ``tensor`` if the observer is enabled.  Then
        returns a fake-quantized version of ``tensor`` if fake quantization is enabled
        and a ``tensor`` unmodified if not.
        """

        if self.is_dynamic:
            self.activation_post_process.reset_parameters()

        if self.is_dynamic or self.is_observer_enabled:
            self.activation_post_process(tensor.detach())

        if not self.is_fake_quant_enabled:
            return tensor

        scale, zero_point = self.activation_post_process.calculate_qparams()
        if self.ch_axis is not None:
            if len(self.ch_axis) > 1:
                new_shape = [
                    i for i in range(len(tensor.shape)) if i not in self.ch_axis
                ] + [i for i in range(len(tensor.shape)) if i in self.ch_axis]
                tensor = tensor.permute(new_shape)
            else:
                tensor = tensor.transpose(self.ch_axis[0], -1)

        out = _tamm_ao_F.fake_quantize(
            tensor,
            scale=scale,
            zero_point=zero_point,
            quant_min=self.activation_post_process.quant_min,
            quant_max=self.activation_post_process.quant_max,
            cast_dtype=self.cast_dtype,
            reciprocal_mul=self.use_reciprocal_mul,
        )
        if self.ch_axis is not None:
            if len(self.ch_axis) > 1:
                out = _tamm_F.inverse_permute(out, new_shape)
            else:
                out = out.transpose(self.ch_axis[0], -1)

        return out


class SimpleEMAMinMaxObserver(_nn.Module):
    """
    Computes the quantization scale parameter based on a moving average of the min
    and max values of recent inputs.

    Args:
        momentum (:obj:`float`, optional): The momentum parameter for the exponential
            moving average of min/max values across batches.  Defaults to ``0.9``.
        quant_min (:obj:`int`, optional): The smallest possible quantized integer value.
            Defaults to ``-128`` for symmetric ``int8`` quantization.
        quant_max (:obj:`int`, optional): The largest possible quantized integer value.
            Defaults to ``127`` for symmetric ``int8`` quantization.
        device (:obj:`torch.devce``, optional): The device for moving average
            parameters.
    """

    def __init__(
        self,
        momentum: float = 0.9,
        quant_min: int = -128,
        quant_max: int = 127,
        device: _Optional[_torch.device] = None,
    ):
        super().__init__()

        if momentum < 0 or momentum >= 1:
            raise ValueError(f"momentum ({momentum}) is not in [0, 1)")
        if quant_max <= 0:
            raise ValueError(f"quant_max ({quant_max}) is not greater than 0")
        if quant_min > 0:
            raise ValueError(f"quant_min ({quant_min}) is greater than 0")

        self.momentum = momentum
        self.quant_min, self.quant_max = quant_min, quant_max

        range_buffer = _torch.empty(2, dtype=_torch.float32, device=device)
        self.register_buffer("range", range_buffer)

        num_samples_buffer = _torch.empty((), dtype=_torch.int64, device=device)
        self.register_buffer("num_samples", num_samples_buffer)

        self.reset_parameters()

    def reset_parameters(self):
        self.range[0] = -1
        self.range[1] = 1
        self.num_samples.zero_()

    def forward(self, input):  # pylint: disable=redefined-builtin
        """
        Updates the moving average of min/max values and returns ``None``.
        """
        self.to(_torch.float32)

        min_max = _torch.stack([input.min(), input.max()]).to(_torch.float32)
        alpha = _torch.tensor(
            1 - self.momentum, dtype=_torch.float32, device=self.range.device
        )
        one = _torch.ones_like(alpha)
        alpha = _torch.where(self.num_samples != 0, alpha, one)
        self.range.add_(min_max - self.range, alpha=alpha)
        self.num_samples.add_(input.size(0) if input.ndim > 0 else 1)

    def calculate_qparams(self):
        """
        Returns ``0`` for the zero point and a scale based on the min/max moving
        average across past batches.  Specifically, the scale is maximum of
        ``min_avg / quant_min`` and ``max_avg / quant_max``.
        """

        self.to(_torch.float32)

        if self.quant_min == 0:
            scale = self.range[1] * (1 / self.quant_max)
        else:
            denom = _torch.tensor(
                [1 / self.quant_min, 1 / self.quant_max], device=self.range.device
            )
            scale = _torch.max(self.range * denom)
        zero_point = _torch.tensor([0], device=self.range.device)
        return scale, zero_point


class SimpleEMAPerChannelMinMaxObserver(_nn.Module):
    """
    Computes the channel-wise quantization scale parameter based on a moving average of
    the min and max values of recent inputs.

    Args:
        ch_axis (:obj:`int` or :obj:`list`): Axis or axes to keep quantization
            parameters independent.
        num_channels (:obj:`int` or :obj:`list`): Expected number of channels,
            corresponding to the axes in ch_axis.
        momentum (:obj:`float`, optional): The momentum parameter for the exponential
            moving average of min/max values across batches.  Defaults to ``0.9``.
        quant_min (:obj:`int`, optional): The smallest possible quantized integer value.
            Defaults to ``-128`` for symmetric ``int8`` quantization.
        quant_max (:obj:`int`, optional): The largest possible quantized integer value.
            Defaults to ``127`` for symmetric ``int8`` quantization.
        device (:obj:`torch.devce``, optional): The device for moving average
            parameters.
    """

    def __init__(
        self,
        ch_axis: _Union[int, list],
        num_channels: _Union[int, list],
        momentum: float = 0.9,
        quant_min: int = -128,
        quant_max: int = 127,
        device: _Optional[_torch.device] = None,
    ):
        super().__init__()

        if isinstance(ch_axis, list):
            assert len(ch_axis) == len(num_channels)
        else:
            assert isinstance(ch_axis, int) and isinstance(num_channels, int)
            ch_axis = [ch_axis]
            num_channels = [num_channels]

        if momentum < 0 or momentum >= 1:
            raise ValueError(f"momentum ({momentum}) is not in [0, 1)")
        if quant_max <= 0:
            raise ValueError(f"quant_max ({quant_max}) is not greater than 0")
        if quant_min > 0:
            raise ValueError(f"quant_min ({quant_min}) is greater than 0")

        self.momentum = momentum
        self.quant_min, self.quant_max = quant_min, quant_max

        range_shape = (*num_channels, 2)

        self.ch_axis = ch_axis
        self.num_channels = num_channels

        range_buffer = _torch.empty(range_shape, dtype=_torch.float32, device=device)
        self.register_buffer("range", range_buffer)

        num_samples_buffer = _torch.empty((), dtype=_torch.int64, device=device)
        self.register_buffer("num_samples", num_samples_buffer)

        self.reset_parameters()

    def reset_parameters(self):
        self.range[..., 0] = -1
        self.range[..., 1] = 1
        self.num_samples.zero_()

    def forward(self, input):  # pylint: disable=redefined-builtin
        """
        Updates the moving average of min/max values and returns ``None``.
        """
        self.to(_torch.float32)

        if len(self.ch_axis) > 1:
            new_shape = [i for i in range(len(input.shape)) if i in self.ch_axis] + [
                i for i in range(len(input.shape)) if i not in self.ch_axis
            ]
            flattened_input = input.permute(tuple(new_shape))
            flattened_input = flattened_input.flatten(start_dim=len(self.ch_axis))
            flattened_input = flattened_input.flatten(end_dim=len(self.ch_axis) - 1)
        else:
            flattened_input = input.transpose(0, self.ch_axis[0]).flatten(start_dim=1)

        min_max = _torch.stack(
            [flattened_input.min(dim=-1)[0], flattened_input.max(dim=-1)[0]], dim=1
        )
        min_max = min_max.to(_torch.float32).reshape(self.range.shape)
        alpha = _torch.tensor(
            1 - self.momentum, dtype=_torch.float32, device=self.range.device
        )
        one = _torch.ones_like(alpha)
        alpha = _torch.where(self.num_samples != 0, alpha, one)
        self.range.add_(min_max - self.range, alpha=alpha)
        self.num_samples.add_(input.size(0) if input.ndim > 0 else 1)

    def calculate_qparams(self):
        """
        Returns ``0`` for the zero point and a scale based on the min/max moving
        average across past batches.  Specifically, the scale is maximum of
        ``min_avg / quant_min`` and ``max_avg / quant_max``.
        """

        self.to(_torch.float32)

        if self.quant_min == 0:
            scale = self.range[..., 1] * (1 / self.quant_max)
        else:
            denom = _torch.tensor(
                [1 / self.quant_min, 1 / self.quant_max], device=self.range.device
            )
            scale = _torch.max(self.range * denom.unsqueeze(0), dim=-1)[0]
        zero_point = _torch.zeros(self.range.shape[:-1], device=self.range.device)
        return scale, zero_point

    def extra_repr(self) -> str:
        return f"ch_axis={self.ch_axis}, num_channels={self.num_channels}"



r"""Bottleneck adapter module."""

# pylint: disable=not-callable, disable=protected-access

import math as _math
from typing import Any as _Any
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn
from torch.nn import functional as _F
from torch.nn import init as _torch_init

from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_adapters.common import (
    attach_config_class as _attach_config_class,
)
from tamm.layers import activation as _activation
from tamm.layers import dropout as _dropout
from tamm.layers import norm as _norm


def _init_weight_value(weight: _nn.Parameter):
    """Initialize the weight parameter's value."""
    _torch_init.kaiming_uniform_(weight, a=_math.sqrt(5))


def init_weight_params(
    output_dim: int,
    input_dim: int,
    device: _Optional[_Union[_torch.device, str]] = None,
    dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
):
    r"""Initialize weight parameters."""
    weight = _nn.Parameter(
        _torch.empty(output_dim, input_dim, device=device, dtype=dtype)
    )
    _init_weight_value(weight)
    return weight


def _init_bias_value(bias: _nn.Parameter, weight: _Optional[_nn.Parameter] = None):
    """Initialize the bias parameter's value."""
    if weight is not None:
        fan_in, _ = _torch_init._calculate_fan_in_and_fan_out(weight)
        bound = 1 / _math.sqrt(fan_in) if fan_in > 0 else 0
        _torch_init.uniform_(bias, -bound, bound)


def init_bias_params(
    dim: int,
    weight: _Optional[_nn.Parameter] = None,
    device: _Optional[_Union[_torch.device, str]] = None,
    dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
):
    r"""Initialize bias parameters."""
    bias = _nn.Parameter(_torch.empty(dim, device=device, dtype=dtype))
    _init_bias_value(bias, weight)
    return bias


@_attach_config_class
class Bottleneck(_LayerAdapter):
    """
    Bottleneck layer adapter.

    Args:
        input_dim: Layer input dim.
        output_dim: Layer output dim.
        hidden_dim: Layer hidden dim.
        activation: Layer activation.
        dropout_p: Dropout rate.
        layer_norm_eps: Layer norm eps.
        layer_norm_bias: Layer norm bias.
        input_normalize: Apply layer norm to inputs.
        device: Torch device type.
        dtype: Tensor dtype.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dim: int = 16,
        a_weights: _Optional[_nn.Module] = None,
        a_bias: _Optional[_nn.Module] = None,
        b_weights: _Optional[_nn.Module] = None,
        b_bias: _Optional[_nn.Module] = None,
        activation: _nn.Module = _activation.ReLU(),
        dropout_p: float = 0.2,
        layer_norm_eps: float = 1e-5,
        layer_norm_bias: bool = True,
        input_normalize: bool = True,
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ):
        super().__init__()

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dim = hidden_dim
        self.a_weights = a_weights
        self.a_bias = a_bias
        self.b_weights = b_weights
        self.b_bias = b_bias
        self.dropout_p = dropout_p
        self.activation = activation
        self.device = device
        self.dtype = dtype

        self.dropout = _dropout.Dropout(p=self.dropout_p)
        self._init_adapter_params()

        self.input_normalize = input_normalize
        if self.input_normalize:
            self.norm = _norm.LayerNorm(
                [self.input_dim],
                eps=layer_norm_eps,
                bias=layer_norm_bias,
                device=device,
                dtype=dtype,
            )

    def _init_adapter_params(self):
        if self.a_weights is None:
            self.a_weights = init_weight_params(
                self.hidden_dim, self.input_dim, device=self.device, dtype=self.dtype
            )

        if self.a_bias is None:
            self.a_bias = init_bias_params(
                self.hidden_dim, self.a_weights, device=self.device, dtype=self.dtype
            )

        if self.b_weights is None:
            self.b_weights = init_weight_params(
                self.output_dim, self.hidden_dim, device=self.device, dtype=self.dtype
            )

        if self.b_bias is None:
            self.b_bias = init_bias_params(
                self.output_dim, self.b_weights, device=self.device, dtype=self.dtype
            )

    def reset_parameters(self):
        super().reset_parameters()
        _init_weight_value(self.a_weights)
        _init_bias_value(self.a_bias, self.a_weights)
        _init_weight_value(self.b_weights)
        _init_bias_value(self.b_bias, self.b_weights)

    def _transform_outputs(
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _torch.Tensor,
    ):
        residual = outputs
        if self.input_normalize:
            outputs = self.norm(outputs)
        outputs = _F.linear(outputs, self.a_weights, self.a_bias)
        outputs = self.dropout(self.activation(outputs))
        outputs = _F.linear(outputs, self.b_weights, self.b_bias)
        return outputs + residual

    def extra_repr(self) -> str:
        return (
            f"out_features={self.output_dim}, "
            f"in_features={self.input_dim}, "
            f"hidden_dim={self.hidden_dim}, "
            f"dropout_p={self.dropout_p}, "
            f"input_normalize={self.input_normalize}"
        )



"""
Builder Post Hooks
------------------
.. autofunction:: tamm.layers.common.get_model_adapters_post_hook

.. autoclass:: tamm.layers.common.AttachConfigPostHook
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.AttachMetadataPostHook
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.FreezeParamsPostHook
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.ArchOptimizersPostHook
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.ModelInitializerPostHook
    :members:
    :show-inheritance:
"""
from tamm.layers.common.post_hooks.adapters import get_model_adapters_post_hook
from tamm.layers.common.post_hooks.arch_optimizer import ArchOptimizersPostHook
from tamm.layers.common.post_hooks.attach_config import AttachConfigPostHook
from tamm.layers.common.post_hooks.attach_metadata import AttachMetadataPostHook
from tamm.layers.common.post_hooks.common import CompositePostHook, IdentityPostHook
from tamm.layers.common.post_hooks.freeze_params import FreezeParamsPostHook
from tamm.layers.common.post_hooks.model_initializer import ModelInitializerPostHook

__all__ = [
    "get_model_adapters_post_hook",
    "ArchOptimizersPostHook",
    "FreezeParamsPostHook",
    "AttachConfigPostHook",
    "AttachMetadataPostHook",
    "ModelInitializerPostHook",
    "IdentityPostHook",
    "CompositePostHook",
]



"""
utils.json
==========

This submodule includes functions for serializing |tamm| objects to JSON.

Example:

To deserialize a model config from JSON and to use it for model creation:

.. tip::
    See :ref:`configure tamm models <configure_models>` for extra context regarding tamm
    model lifecycle.

.. code-block:: python

    import tamm.utils.json as tamm_json

    with open("invalid_type.json", "r", encoding="utf-8") as fptr:
        config = tamm_json.load(fptr)

    model = config.create_model()


To serialize a model config from JSON and to use it for model creation:

.. code-block:: python

    import tamm
    import tamm.utils.json as tamm_json

    config = tamm.create_model_config(model_id)
    # apply customizations to config (e.g., change number of layers)
    config.num_layers = 8
    with open("customized_model_config.json", "w", encoding="utf-8") as fptr:
        tamm_json.dump(config, fptr)

.. autofunction:: load
.. autofunction:: loads
.. autofunction:: dump
.. autofunction:: dumps

Classes useful for plugin developers
------------------------------------

.. autoclass:: tamm.utils.json.JSONSerializableMixin
   :members:
   :private-members:

.. autoclass:: tamm.utils.json.JSONSerializableABCMixin

.. autoclass:: tamm.utils.json.SupportsObjectHook
    :members:

.. autofunction:: tamm.utils.json.register_load_normalizer

.. autoclass:: tamm.utils.json.BackwardCompatibilityTypeNormalizer
    :show-inheritance:
"""

from tamm.utils.json._normalization import (
    BackwardCompatibilityTypeNormalizer,
    SupportsObjectHook,
    register_load_normalizer,
)
from tamm.utils.json._utils import JSONSerializableABCMixin, JSONSerializableMixin
from tamm.utils.json.api import dump, dumps, load, loads

__all__ = [
    "BackwardCompatibilityTypeNormalizer",
    "dump",
    "dumps",
    "load",
    "loads",
    "JSONSerializableABCMixin",
    "JSONSerializableMixin",
    "SupportsObjectHook",
    "register_load_normalizer",
]



"""
transformer.vision
^^^^^^^^^^^^^^^^^^

This submodule implements transformer-specific components for vision models.

.. autoclass:: tamm.layers.VisionTransformerEncoder
    :show-inheritance:
    :members: forward
"""

# pylint: disable=duplicate-code

from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch

from tamm import _helpers
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.transformer import stack as _transformer_stack
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder
from tamm.typing import OptionalTensor as _OptionalTensor


class VisionTransformerEncoder(
    _transformer_stack.ExtendedTransformerStack, _LayerMixin
):
    """
    A general layer for vision encoders.  This layer is the same as
    :class:`.ExtendedTransformerStack` except for a few changes:

    * The layer has an optional ``segmentation`` layer, and the :meth:`forward` method
      accepts an optional ``padding_mask`` argument.  The segmentation layer computes
      default ``segment_ids`` from the ``padding_mask``.
    * The :meth:`forward` method does not take arguments for KV caching or cross
      attention.

    Args:
        segmentation: A layer for converting per-pixel ``padding_mask`` inputs to
            per-token segment IDs with shape ``(batch_size, *seq_shape)``.  Typically
            the segment IDs also form a padding mask with ``0`` for padding values and
            ``1`` for non-padding.

    Please see :class:`.ExtendedTransformerStack` for a description of the remaining
    parameters.
    """

    def __init__(
        self,
        *,
        segmentation: _OptionalModuleOrBuilder = None,
        embedding: _OptionalModuleOrBuilder = None,
        token_metadata_logic: _OptionalModuleOrBuilder = None,
        token_type_encoding: _OptionalModuleOrBuilder = None,
        attention_mask: _OptionalModuleOrBuilder = None,
        positional_encoding: _OptionalModuleOrBuilder = None,
        secondary_positional_encodings: _OptionalModuleOrBuilder = None,
        input_norm: _OptionalModuleOrBuilder = None,
        layers: _ModuleOrBuilder,
        output_norm: _OptionalModuleOrBuilder = None,
        output_transform: _OptionalModuleOrBuilder = None,
        output_transform_build_mode: _Optional[str] = "basic",
        loss: _OptionalModuleOrBuilder = None,
        embeddings_cast_dtype: _Optional[_Union[str, _torch.dtype]] = "auto",
    ):
        super().__init__(
            embedding=embedding,
            token_metadata_logic=token_metadata_logic,
            token_type_encoding=token_type_encoding,
            attention_mask=attention_mask,
            positional_encoding=positional_encoding,
            secondary_positional_encodings=secondary_positional_encodings,
            input_norm=input_norm,
            layers=layers,
            output_norm=output_norm,
            output_transform=output_transform,
            output_transform_build_mode=output_transform_build_mode,
            loss=loss,
            embeddings_cast_dtype=embeddings_cast_dtype,
        )
        _helpers.prepend_children(self, segmentation=segmentation)

    def forward(  # pylint: disable=arguments-differ
        self,
        inputs: _Any,
        *,
        labels: _Any = None,
        padding_mask: _Any = None,
        segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        attention_mask: _OptionalTensor = None,
        aux_token_metadata: _Optional[_Dict[str, _Any]] = None,
        mode: _Union[_transformer_stack.ExtendedTransformerStackMode, str] = "full",
    ) -> _transformer_stack.ExtendedTransformerStackOutput:
        """
        Args:

            inputs (:obj:`torch.Tensor`): Inputs to the ``embedding`` layer, which
                typically are images in NCHW format.  For some values of ``mode``, this
                should instead be input embeddings.
            labels: Optional labels for the ``loss`` layer.  If provided, the loss layer
                should accept two positional arguments; the first is the output of
                ``output_transform``, and the second is ``labels``.
            padding_mask (:obj:`torch.Tensor`, optional): An optional padding mask to
                pass to the ``segmentation`` layer.  Typically this is a pixel-level
                mask.  Defaults to ``None``.
            segment_ids (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that controls the attention mask for
                self attention layers.  For sequence ``i``, if
                ``segment_ids[i, j] != segment_ids[i, k]``, then tokens ``j`` and
                ``k`` do not attend to each other.  If unspecified, the ``segmentation``
                layer computes ``segment_ids`` from the ``padding_mask``.  If the
                ``segmentation`` layer or ``padding_mask`` is not provided, then
                ``segment_ids`` defaults to all ones.
            positions (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies the position of each
                token.  If unspecified, ``positions`` defaults to
                ``0, 1, 2, ...`` for each segment of each sequence.
            token_types (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies a "type" for each token.
                The ``token_type_encoding`` and ``attention_mask`` child layers may use
                this type information in different ways, depending on the model.
            attention_mask (:obj:`torch.Tensor`, optional): A boolean tensor with shape
                ``(batch_size, seq_len, seq_len)``.  A ``True`` value indicates that
                the corresponding pair of tokens should take part in attention.
            aux_token_metadata (:obj:`dict`): A dictionary of optional auxiliary inputs for
                extending model behavior.  This gets attached to the ``token_metadata``
                input of several child layers, including the attention mask, layer
                sequence, and positional encoding layers.
            mode (:obj:`str` or :obj:`.ExtendedTransformerStackMode`, optional): A
                mode for controlling the forward behavior.  Defaults to ``FULL``.
        """

        mode = _helpers.get_enum_member_from_name(
            _transformer_stack.ExtendedTransformerStackMode, mode
        )
        segment_ids = self._maybe_update_segment_ids(
            segment_ids, padding_mask=padding_mask
        )
        return super().forward(
            inputs=inputs,
            labels=labels,
            segment_ids=segment_ids,
            positions=positions,
            token_types=token_types,
            attention_mask=attention_mask,
            aux_token_metadata=aux_token_metadata,
            mode=mode,
        )

    def _maybe_update_segment_ids(self, segment_ids, *, padding_mask):
        if segment_ids is not None:
            return segment_ids
        if self.segmentation is not None and padding_mask is not None:
            return self.segmentation(padding_mask)
        return None



from tamm.model_repo import (
    create_adapted_model,
    create_model,
    create_model_builder,
    create_model_config,
    is_adapted_model_name,
    is_model_builder_name,
    is_model_config_name,
    is_model_name,
    list_adapted_models,
    list_model_builders,
    list_model_configs,
    list_models,
)
from tamm.models import afm_text, common
from tamm.models.vision_transformer import VisionTransformer

__all__ = [
    "afm_text",
    "common",
    "create_model",
    "create_adapted_model",
    "create_model_builder",
    "create_model_config",
    "list_adapted_models",
    "list_model_builders",
    "list_model_configs",
    "list_models",
    "is_model_name",
    "is_adapted_model_name",
    "is_model_builder_name",
    "is_model_config_name",
]
